import ntpath

import cudf

from optimus.engines.base.io.load import BaseLoad
from optimus.engines.base.meta import Meta
from optimus.engines.cudf.dataframe import CUDFDataFrame
from optimus.helpers.functions import prepare_path, unquote_path
from optimus.helpers.logger import logger


class Load(BaseLoad):

    def __init__(self, op):
        self.op = op

    @staticmethod
    def json(path, multiline=False, n_rows=-1, *args, **kwargs):
        """
        Return a dataframe from a json file.
        :param path: path or location of the file.
        :param multiline:

        :return:
        """

        path = unquote_path(path)

        local_file_names = prepare_path(path, "json")
        try:
            df_list = []

            for file_name, j in local_file_names:
                df = cudf.read_json(file_name, lines=multiline, nrows=n_rows, *args, **kwargs)
                df_list.append(df)

            df = cudf.concat(df_list, axis=0, ignore_index=True)
            df = CUDFDataFrame(df)
            df.meta = Meta.set(df.meta, "file_name", local_file_names[0])

        except IOError as error:
            logger.print(error)
            raise
        return df

    @staticmethod
    def tsv(path, header=True, infer_schema=True, *args, **kwargs):
        """
        Return a dataframe from a tsv file.
        :param path: path or location of the file.
        :param header: tell the function whether dataset has a header row. True default.
        :param infer_schema: infers the input schema automatically from data.
        It requires one extra pass over the data. True default.

        :return:
        """

        return Load.csv(path, sep='\t', header=header, infer_schema=infer_schema, *args, **kwargs)

    @staticmethod
    def csv(path, sep=',', header=True, infer_schema=True, encoding="utf-8", null_value="None", n_rows=-1, cache=False,
            quoting=0, lineterminator=None, error_bad_lines=False, keep_default_na=False, na_filter=True, dtype=None,
            *args, **kwargs):

        """
        Return a dataframe from a csv file.
        params

        :param dtype:
        :param cache:
        :param na_filter:
        :param path: path or location of the file.
        :param sep: usually delimiter mark are ',' or ';'.
        :param keep_default_na:
        :param error_bad_lines:
        :param quoting:
        :param lineterminator:
        :param header: tell the function whether dataset has a header row. True default.
        :param infer_schema: infers the input schema automatically from data.
        :param null_value:
        :param n_rows:
        :param encoding:
        It requires one extra pass over the data. True default.

        :return dataFrame
        """

        path = unquote_path(path)

        # file, file_name = prepare_path(path, "csv")[0]

        try:
            # TODO:  lineterminator=lineterminator seems to be broken
            if header is True:
                header = 0
            elif header is False:
                header = None
            # The str to ["str] is due to a bug in cudf https://github.com/rapidsai/cudf/issues/6606
            if dtype == str or dtype is None:
                dtype = ["str"]

            cdf = cudf.read_csv(path, sep=sep, header=header, encoding=encoding,
                                quoting=quoting, error_bad_lines=error_bad_lines,
                                keep_default_na=keep_default_na, na_values=null_value, nrows=n_rows,
                                na_filter=na_filter, dtype=dtype, *args, **kwargs)
            df = CUDFDataFrame(cdf)
            df.meta = Meta.set(df.meta, None,
                               {"file_name": path, "max_cell_length": df.cols.len("*").cols.max()})


        except IOError as error:
            logger.print(error)
            raise
        return df

    @staticmethod
    def parquet(path, columns=None, *args, **kwargs):
        """
        Return a dataframe from a parquet file.
        :param path: path or location of the file. Must be string dataType
        :param columns: select the columns that will be loaded. In this way you do not need to load all the dataframe
        :param args: custom argument to be passed to the parquet function
        :param kwargs: custom keyword arguments to be passed to the parquet function
        """

        path = unquote_path(path)

        try:
            df = cudf.read_parquet(path, columns=columns, engine='pyarrow', *args, **kwargs)
            df = CUDFDataFrame(df)
            df.meta = Meta.set(df.meta, "file_name", path)

        except IOError as error:
            logger.print(error)
            raise

        return df

    @staticmethod
    def avro(path, storage_options=None, conn=None, *args, **kwargs):
        """
        :param path: path or location of the file. Must be string dataType
        :param storage_options:
        :param args: custom argument to be passed to the avro function
        :param kwargs: custom keyword arguments to be passed to the avro function
        """

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "avro")[0]

        try:
            df = cudf.read_avro(path, storage_options=storage_options, *args, **kwargs)
            df = CUDFDataFrame(df)
            df.meta = Meta.set(df.meta, "file_name", file_name)

        except IOError as error:
            logger.print(error)
            raise

        return df

    @staticmethod
    def orc(path, columns=None, storage_options=None, conn=None, *args, **kwargs):
        """
        Return a dataframe from a avro file.
        :param path: path or location of the file. Must be string dataType
        :param columns: Subset of columns to be loaded
        :param args: custom argument to be passed to the avro function
        :param kwargs: custom keyword arguments to be passed to the avro function
        """

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "orc")[0]

        try:
            df = cudf.read_orc(path, columns, storage_options=storage_options, *args, **kwargs)
            df = CUDFDataFrame(df)
            df.meta = Meta.set(df.meta, "file_name", file_name)

        except IOError as error:
            logger.print(error)
            raise

        return df

    @staticmethod
    def excel(path, sheet_name=0, storage_options=None, conn=None, *args, **kwargs):
        """
        Return a dataframe from a excel file.
        :param path: Path or location of the file. Must be string dataType
        :param sheet_name: excel sheet name
        :param args: custom argument to be passed to the excel function
        :param kwargs: custom keyword arguments to be passed to the excel function
        """

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "xls")

        try:
            pdf = cudf.read_excel(file, sheet_name=sheet_name, storage_options=storage_options, *args, **kwargs)

            # exception when Spark try to infer the column data type
            # Parse object column data type to string to ensure that Spark can handle it. With this we try to reduce
            col_names = list(pdf.select_dtypes(include=['object']))

            column_dtype = {}
            for col in col_names:
                column_dtype[col] = str

            # Convert object columns to string
            pdf = pdf.astype(column_dtype)

            df = cudf.from_pandas(pdf, npartitions=3)
            df.meta = Meta.set(df.meta, "file_name", ntpath.basename(file_name))
        except IOError as error:
            logger.print(error)
            raise

        return df



import os
from .optimus import optimus as Optimus
from optimus._version import __version__

ROOT_DIR = os.path.dirname(os.path.abspath(
    __file__))  # This is your Project Root


# Preserve compatibility with <3.x branch
os.environ["PYTHONIOENCODING"] = "utf8"



import functools
import operator

import pandas as pd
from multipledispatch import dispatch

from optimus.engines.base.meta import Meta
from optimus.engines.base.rows import BaseRows
from optimus.helpers.columns import parse_columns
from optimus.helpers.constants import Actions
from optimus.helpers.core import val_to_list, one_list_to_val
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_list_of_str_or_int, is_list_value

DataFrame = pd.DataFrame


class Rows(BaseRows):
    def __init__(self, df):
        super(Rows, self).__init__(df)

    @staticmethod
    def create_id(column="id") -> DataFrame:
        pass

    def append(self, rows):
        """

        :param rows:
        :return:
        """
        df = self.root

        if is_list_value(rows):
            rows = self.root.new(pd.DataFrame(rows))
        # Can not concatenate dataframe with not string columns names

        rows.columns = df.cols.names()
        df.data.reset_index(drop=True)
        df = pd.concat([df.data.reset_index(drop=True), rows.data.reset_index(drop=True)], axis=0)
        return self.root.new(df)

    def _sort(self, dfd, col_name, ascending):
        return dfd.sort_values(col_name, ascending=ascending)

    @staticmethod
    def unnest(input_cols) -> DataFrame:
        df = self
        return df



import pandavro as pdx
import os
from optimus.helpers.logger import logger


class Save:
    def __init__(self, root):
        self.root = root

    def json(self, path, orient="records", *args, **kwargs):
        """
        Save data frame in a json file
        :param path: path where the spark will be saved.
        :param orient:

        :return:
        """
        df = self.root.data

        try:
            df.to_json(path, orient=orient, *args, **kwargs)

        except IOError as e:
            logger.print(e)
            raise

    def csv(self, path, mode="w",show_path=False, **kwargs):
        """
        Save data frame to a CSV file.
        :param path: path where the spark will be saved.
        :param mode: 'rb', 'wt', etc
        it uses the default value.
        :return: Dataframe in a CSV format in the specified path.
        """

        try:
            df = self.root.data
            # columns = parse_columns(self, "*",
            #                         filter_by_column_dtypes=["date", "array", "vector", "binary", "null"])
            # df = df.cols.cast(columns, "str").repartition(num_partitions)

            # Dask reference
            # https://docs.dask.org/en/latest/dataframe-api.html#dask.dataframe.to_csv
            # df.to_csv(filename=path)

            if show_path is True:
                print(os.path.abspath(path))

            df.to_csv(path, index=False, mode=mode)

        except IOError as error:
            logger.print(error)
            raise

    def excel(self, path, **kwargs):
        """
        Save data frame to a CSV file.
        :param path: path where the spark will be saved.
        :param mode: 'rb', 'wt', etc
        it uses the default value.
        :return: Dataframe in a CSV format in the specified path.
        """

        try:
            df = self.root.data
            df.to_excel(path, index=False)

        except IOError as error:
            logger.print(error)
            raise

    def xml(self, filename=None, mode='w'):
        df = self.root.data

        def row_to_xml(row):
            xml = ['<item>']
            for i, col_name in enumerate(row.index):
                xml.append('  <field name="{0}">{1}</field>'.format(col_name, row.iloc[i]))
            xml.append('</item>')
            return '\n'.join(xml)

        res = '\n'.join(df.apply(row_to_xml, axis=1))

        if filename is None:
            return res
        with open(filename, mode) as f:
            f.write(res)

    def parquet(self, path, *args, **kwargs):
        """
        Save data frame to a parquet file
        :param path: path where the spark will be saved.
        :param mode: Specifies the behavior of the save operation when data already exists.
                    "append": Append contents of this DataFrame to existing data.
                    "overwrite" (default case): Overwrite existing data.
                    "ignore": Silently ignore this operation if data already exists.
                    "error": Throw an exception if data already exists.
        :param num_partitions: the number of partitions of the DataFrame
        :return:
        """

        # This character are invalid as column names by parquet
        invalid_character = [" ", ",", ";", "{", "}", "(", ")", "\n", "\t", "="]

        def func(col_name):
            for i in invalid_character:
                col_name = col_name.replace(i, "_")
            return col_name

        df = self.root.cols.rename(func)

        try:
            df.data.to_parquet(path)
        except IOError as e:
            logger.print(e)
            raise

    def avro(self, path):
        pdx.to_avro(path, self.root.data)



from functools import reduce

from optimus.helpers.core import val_to_list
from optimus.helpers.functions import random_int
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_

def traverse(obj, path=None, callback=None):
    """
    Traverse a deep nested python structure
    :param obj: object to traverse
    :param path:
    :param callback: Function used to transform a value
    :return:
    """
    from pyspark.ml.linalg import DenseVector

    if path is None:
        path = []

    if is_(obj, dict):
        value = {k: traverse(v, path + [k], callback)
                 for k, v in obj.items()}

    elif is_(obj, list):
        value = [traverse(elem, path + [[]], callback)
                 for elem in obj]

    elif is_(obj, tuple):
        value = tuple(traverse(elem, path + [[]], callback)
                      for elem in obj)
    elif is_(obj, DenseVector):
        value = DenseVector([traverse(elem, path + [[]], callback) for elem in obj])
    else:
        value = obj

    if callback is None:  # if a callback is provided, call it to get the new value
        return value
    else:
        return callback(path, value)


def append(dfs, like="columns"):
    """
    Concat multiple dataFrames columns or rows wise
    :param dfs: List of DataFrames
    :param like: concat as columns or rows
    :return:
    """

    # FIX: Because monotonically_increasing_id can create different
    # sequence for different dataframes the result could be wrong.

    if like == "columns":
        temp_dfs = []
        col_temp_name = "id_" + random_int()

        dfs = val_to_list(dfs)
        for df in dfs:
            from pyspark.sql import functions as F
            temp_dfs.append(df.withColumn(col_temp_name, F.monotonically_increasing_id()))

        def _append(df1, df2):
            return df1.join(df2, col_temp_name, "outer")

        df_result = reduce(_append, temp_dfs).drop(col_temp_name)

    elif like == "rows":
        from pyspark.sql import DataFrame
        df_result = reduce(DataFrame.union, dfs)
    else:
        RaiseIt.value_error(like, ["columns", "rows"])

    return df_result


from typing import TypeVar, List

DataFrameType = TypeVar("DataFrameType")
InternalDataFrameType = TypeVar("InternalDataFrameType")
MaskDataFrameType = TypeVar("MaskDataFrameType")
ConnectionType = TypeVar("ConnectionType")
ClustersType = TypeVar("ClustersType")
ModelType = TypeVar("ModelType")

DataFrameTypeList = TypeVar("DataFrameTypeList")
InternalDataFrameTypeList = TypeVar("InternalDataFrameTypeList")
MaskDataFrameTypeList = TypeVar("MaskDataFrameTypeList")
ConnectionTypeList = TypeVar("ConnectionTypeList")
ClustersTypeList = TypeVar("ClustersTypeList")
ModelTypeList = TypeVar("ModelTypeList")

StringsList = TypeVar("StringsList", List[str], str)
StringsListNone = TypeVar("StringsListNone", List[str], str, None)

_list_types = [str(DataFrameTypeList), str(InternalDataFrameTypeList), str(MaskDataFrameTypeList), str(ConnectionTypeList), str(ClustersTypeList), str(ModelTypeList)]

_types = [str(DataFrameType), str(InternalDataFrameType), str(MaskDataFrameType), str(ConnectionType), str(ClustersType), str(ModelType)]

def is_list_of_optimus_type(value):
    return str(value) in _list_types

def is_optimus_type(value):
    return str(value) in _types

def is_any_optimus_type(value):
    return str(value) in [*_types, *_list_types]


import numpy as np

class BaseConstants:

    DTYPES_ALIAS = {"str": "string",
                "vector": "object",
                "list": "object",
                "dict": "object",
                "datetime": "datetime64[ns]",
                "date": "datetime64[ns]",
                "time": "datetime64[ns]",
                "timestamp": "datetime64[ns]"}

    DTYPES_DICT = {"string": np.str, "uint8": np.uint8, "uint16": np.uint16, "uint32": np.uint32,
                    "uint64": np.uint64, "int8": np.int8, "int16": np.int16, "int32": np.int32, "int64": np.int64,
                    "float": np.float, "float64": np.float64, "boolean": np.bool, "array": np.array,
                    "bigint": np.int64, "object": np.object_}

    SHORT_DTYPES = {"string": "string",
                    "str": "string",
                    "integer": "int",
                    "int": "int",

                    "uint8": "uint8",
                    "uint16": "uint16",
                    "uint32": "uint32",
                    "uint64": "uint64",
                    "int8": "int8",
                    "int16": "int16",
                    "int32": "int32",
                    "int64": "int64",

                    "float64": "float64",
                    "big": "bigint",
                    "long": "bigint",
                    "float": "float",
                    "double": "double",
                    "bool": "boolean",
                    "boolean": "boolean",
                    "struct": "struct",
                    "array": "array",
                    "date": "date",
                    "datetime": "datetime",
                    "byte": "byte",
                    "short": "short",
                    "binary": "binary",
                    "null": "null",
                    "vector": "vector",
                    "timestamp": "datetime",
                    "object": "object"
                    }

    NUMERIC_TYPES = ["int8", "int16", "int32", "int64", "uint8", "uint16", "uint32", "uint64", "float64"]
    DTYPES_TO_PROFILER = {"int": ["int64", "int32"], "float": ["float64", "float"], "object": ["object"]}

    STRING_TYPES = ["string", "object"]
    OBJECT_TYPES = ["object"]

LIMIT = 1000
LIMIT_TABLE = 10
NUM_PARTITIONS = 10
SAMPLE_NUMBER = 10000


import re
from abc import abstractmethod, ABC

import jellyfish
import numpy as np
import pandas as pd
from jsonschema._format import is_email
import url_parser

from optimus.helpers.constants import ProfilerDataTypes
from optimus.helpers.core import val_to_list
from optimus.infer import is_list, is_null, is_bool, \
    is_credit_card_number, is_zip_code, is_int, is_decimal, is_datetime, is_object_value, is_ip, is_url, is_missing, \
    is_gender, is_list_of_int, is_list_of_str, is_str, is_phone_number, is_int_like


# ^(?:(?P<protocol>[\w\d]+)(?:\:\/\/))?(?P<sub_domain>(?P<www>(?:www)?)(?:\.?)(?:(?:[\w\d-]+|\.)*?)?)(?:\.?)(?P<domain>[^./]+(?=\.))\.(?P<top_domain>com(?![^/|:?#]))?(?P<port>(:)(\d+))?(?P<path>(?P<dir>\/(?:[^/\r\n]+(?:/))+)?(?:\/?)(?P<file>[^?#\r\n]+)?)?(?:\#(?P<fragment>[^#?\r\n]*))?(?:\?(?P<query>.*(?=$)))*$

class Functions(ABC):
    @staticmethod
    def delayed(func):
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    def from_delayed(self, delayed):
        return delayed[0]

    def to_delayed(self, delayed):
        return [delayed]

    # def _to_float(self, series):
    #     return series

    def to_integer(self, series):
        return series

    def _to_float_partition(self, series):
        return self.to_float(series)

    def _to_integer_partition(self, series):
        return self.to_integer(series)

    def to_string(self, series):
        return series

    def to_string_accessor(self, series):
        return self.to_string(series).str

    # Aggregation
    def min(self, series):
        return series.min()

    def max(self, series):
        return series.max()

    def mean(self, series):
        return self.to_float(series).mean()

    def mode(self, series):
        return self.to_float(series).mode().to_dict()

    def std(self, series):
        return self.to_float(series).std()

    def sum(self, series):
        return self.to_float(series).sum()

    def cumsum(self, series):
        return self.to_float(series).cumsum()

    def cumprod(self, series):
        return self.to_float(series).cumprod()

    def cummax(self, series):
        return self.to_float(series).cummax()

    def cummin(self, series):
        return self.to_float(series).cummin()

    def var(self, series):
        return self.to_float(series).var()

    def count_uniques(self, series, values=None, estimate: bool = True):
        return self.to_string(series).nunique()

    def unique_values(self, series, *args):
        # Cudf can not handle null so we fill it with non zero values.
        return list(self.to_string(series).unique())

    @staticmethod
    def count_na(series):
        return series.isna().sum()
        # return {"count_na": {col_name:  for col_name in columns}}
        # return np.count_nonzero(_df[_serie].isnull().values.ravel())
        # return cp.count_nonzero(_df[_serie].isnull().values.ravel())

    @staticmethod
    @abstractmethod
    def count_zeros(series, *args):
        pass

    @staticmethod
    @abstractmethod
    def kurtosis(series):
        pass

    @staticmethod
    @abstractmethod
    def skew(series):
        pass

    def mad(self, series, error, more):
        series = self.to_float(series)
        series = series[series.notnull()]
        median_value = series.quantile(0.5)
        mad_value = {"mad": (series - median_value).abs().quantile(0.5)}
        if more:
            mad_value.update({"median": median_value})

        return mad_value

    # TODO: dask seems more efficient triggering multiple .min() task, one for every column
    # cudf seems to be calculate faster in on pass using df.min()
    def range(self, series):

        return {"min": self.to_float(series).min(), "max": self.to_float(series).max()}

    def var(self, series):
        return self.to_float(series).var()

    def percentile(self, series, values, error):

        series = self.to_float(series)

        @self.delayed
        def to_dict(_result):
            ## In pandas if all values are none it return {} on dict
            # Dask raise an exception is all values in the series are np.nan
            if _result.isnull().all():
                return np.nan
            else:
                return _result.quantile(values).to_dict()

        return to_dict(series)

    # def radians(series):
    #     return series.to_float().radians()
    #
    # def degrees(series, *args):
    #     return call(series, method_name="degrees")

    ###########################

    def z_score(self, series):
        t = self.to_float(series)
        return t - t.mean() / t.std(ddof=0)

    def modified_z_score(self, series):
        mad_median = self.mad(series, True, True)
        median = mad_median["median"]
        mad = mad_median["mad"]

        return abs(0.6745 * (series - median) / mad)

    def clip(self, series, lower_bound, upper_bound):
        return self.to_float(series).clip(float(lower_bound), float(upper_bound))

    def cut(self, series, bins, labels, default):
        if is_list_of_int(bins):
            return pd.cut(self.to_float(series), bins, include_lowest=True, labels=labels)
        elif is_list_of_str(bins):
            conditions = [series.str.contains(i) for i in bins]

            return np.select(conditions, labels, default=default)

    def qcut(self, series, quantiles):
        return pd.qcut(series, quantiles)

    def abs(self, series):
        return self.to_float(series).abs()

    def exp(self, series):
        return self.to_float(series).exp()

    @abstractmethod
    def word_tokenize(self, series):
        pass

    def len(self, series):
        return self.to_string_accessor(series).len()

    @staticmethod
    @abstractmethod
    def sqrt(series):
        pass

    def mod(self, series, other):
        return self.to_float(series).mod(other)

    def round(self, series, decimals):
        return self.to_float(series).round(decimals)

    def pow(self, series, exponent):
        return self.to_float(series).pow(exponent)

    def floor(self, series):
        return self.to_float(series).floor()

    # def trunc(self):
    #     series = self.series
    #     return series.to_float().truncate()

    @staticmethod
    @abstractmethod
    def radians(series):
        pass

    @staticmethod
    @abstractmethod
    def degrees(series):
        pass

    @staticmethod
    @abstractmethod
    def ln(series):
        pass

    @staticmethod
    @abstractmethod
    def log(series, base):
        pass

    @staticmethod
    @abstractmethod
    def ceil(series):
        pass

    @abstractmethod
    def sin(self, series):
        pass

    @abstractmethod
    def cos(self, series):
        pass

    @abstractmethod
    def tan(self, series):
        pass

    @abstractmethod
    def asin(self, series):
        pass

    @abstractmethod
    def acos(self, series):
        pass

    @abstractmethod
    def atan(self, series):
        pass

    @abstractmethod
    def sinh(self, series):
        pass

    @abstractmethod
    def cosh(self, series):
        pass

    @abstractmethod
    def tanh(self, series):
        pass

    @abstractmethod
    def asinh(self, series):
        pass

    @abstractmethod
    def acosh(self, series):
        pass

    @abstractmethod
    def atanh(self, series):
        pass

    # Strings
    def match(self, series, regex):
        return self.to_string_accessor(series).match(regex)

    def lower(self, series):
        return self.to_string_accessor(series).lower()

    def upper(self, series):
        return self.to_string_accessor(series).upper()

    def title(self, series):
        return self.to_string_accessor(series).title()

    def capitalize(self, series):
        return self.to_string_accessor(series).capitalize()

    def pad(self, series, width, side, fillchar=""):
        return self.to_string_accessor(series).pad(width, side, fillchar)

    def extract(self, series, regex):
        return self.to_string_accessor(series).extract(regex)

    def slice(self, series, start, stop, step):
        return self.to_string_accessor(series).slice(start, stop, step)

    def proper(self, series):
        return self.to_string_accessor(series).title()

    def trim(self, series):
        return self.to_string_accessor(series).strip()

    def strip_html(self, value):
        return re.sub('<.*?>', '', value)

    @staticmethod
    @abstractmethod
    def replace_chars(series, search, replace_by):
        pass

    def replace_words(self, series, search, replace_by):
        search = val_to_list(search)
        str_regex = [r'\b%s\b' % re.escape(s) for s in search]
        # str_regex = (r'\b%s\b' % r'\b|\b'.join(map(re.escape, search)))
        return series.replace(str_regex, replace_by, regex= True)

    def replace_full(self, series, search, replace_by):
        search = val_to_list(search)
        str_regex = [r'^%s$' % re.escape(s) for s in search]
        # str_regex = (r'^%s$' % r'$|^'.join(map(re.escape, search)))
        return series.replace(str_regex, replace_by, regex=True)

    def replace_values(self, series, search, replace_by):
        search = val_to_list(search)
        # return series.mask(series.isin(search), replace_by)
        return series.replace(search, replace_by)

    def remove_white_spaces(self, series):
        return self.to_string_accessor(series).replace(" ", "")

    def remove_urls(self, series):
        return self.to_string_accessor(series).replace(r"https?://\S+|www\.\S+", "", regex=True)

    def normalize_spaces(self, series):
        return self.to_string_accessor(series).replace(r" +", " ", regex=True)

    def expand_contractions(self, series):

        pass
    # @staticmethod
    # def len(series):
    #     return series.str.len()

    def to_datetime(self, series, format):
        pass

    def normalize_chars(self, series):
        pass

    def find(self, sub, start=0, end=None):
        series = self.series
        return self.to_string_accessor(series).find(sub, start, end)

    def rfind(self, series, sub, start=0, end=None):
        return self.to_string_accessor(series).rfind(sub, start, end)

    def left(self, series, position):
        return self.to_string_accessor(series)[:position]

    def right(self, series, position):
        return self.to_string_accessor(series)[-1 * position:]

    def mid(self, series, _start, _n):
        return self.to_string_accessor(series)[_start:_n]

    def starts_with(self, series, pat):
        return self.to_string_accessor(series).startswith(pat)

    def ends_with(self, series, pat):
        return self.to_string_accessor(series).endswith(pat)

    def contains(self, series, pat):
        return self.to_string_accessor(series).contains(pat)

    def char(self, series, _n):
        return self.to_string_accessor(series)[_n]

    @staticmethod
    def unicode(series):
        pass

    @staticmethod
    def exact(series, pat):
        return series == pat

    # dates
    def year(self, series, format):
        """
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        # return self.to_datetime(format=format).strftime('%Y').to_self().reset_index(drop=True)
        return self.to_datetime(series, format=format).dt.year

    @staticmethod
    def month(series, format):
        return series.to_datetime(format=format).dt.month

    @staticmethod
    def day(series, format):
        return series.to_datetime(format=format).dt.day

    @staticmethod
    def hour(series, format):
        return series.to_datetime(format=format).dt.hour

    @staticmethod
    def minute(series, format):
        return series.to_datetime(format=format).dt.minute

    @staticmethod
    def second(series, format):
        return series.to_datetime(format=format).dt.second

    @staticmethod
    def weekday(series, format):
        return series.to_datetime(format=format).dt.weekday

    @staticmethod
    @abstractmethod
    def date_format(self, current_format=None, output_format=None):
        pass

    @staticmethod
    @abstractmethod
    def days_between(self, date_format=None):
        pass

    def domain(self, value):
        return url_parser.parse_url(value)["domain"]

    def top_domain(self, value):
        return url_parser.parse_url(value)["top_domain"]

    def sub_domain(self, value):
        return url_parser.parse_url(value)["sub_domain"]

    def url_scheme(self, value):
        return url_parser.parse_url(value)["protocol"]

    def url_path(self, value):
        return url_parser.parse_url(value)["path"]

    def url_file(self, value):
        return url_parser.parse_url(value)["file"]

    def url_query(self, value):
        return url_parser.parse_url(value)["query"]

    def url_fragment(self, value):
        return url_parser.parse_url(value)["fragment"]

    def host(self, value):
        return url_parser.parse_url(value)["host"]

    def port(self, value):
        return url_parser.parse_url(value)["port"]

    def email_username(self, series):
        return series.str.split('@').str[0]

    def email_domain(self, series):
        return series.str.split('@').str[1]

    def infer_dtypes(self, value, cols_dtype):
        """
        Infer the data types.
        Please be aware that the order in which the value is checked is important and will change the final result
        :param value:
        :param cols_dtype:
        :return:
        """

        if is_list(value):
            dtype = ProfilerDataTypes.ARRAY.value
        elif is_null(value):
            dtype = ProfilerDataTypes.NULL.value
        elif is_bool(value):
            dtype = ProfilerDataTypes.BOOLEAN.value
        elif is_credit_card_number(value):
            dtype = ProfilerDataTypes.CREDIT_CARD_NUMBER.value
        elif is_zip_code(value):
            dtype = ProfilerDataTypes.ZIP_CODE.value
        elif is_int_like(value):
            dtype = ProfilerDataTypes.INT.value
        elif is_decimal(value):
            dtype = ProfilerDataTypes.DECIMAL.value
        elif is_datetime(value):
            dtype = ProfilerDataTypes.DATETIME.value
        elif is_missing(value):
            dtype = ProfilerDataTypes.MISSING.value
        elif is_str(value):
            if is_ip(value):
                dtype = ProfilerDataTypes.IP.value
            elif is_url(value):
                dtype = ProfilerDataTypes.URL.value
            elif is_email(value):
                dtype = ProfilerDataTypes.EMAIL.value
            elif is_gender(value):
                dtype = ProfilerDataTypes.GENDER.value
            elif is_phone_number(value):
                dtype = ProfilerDataTypes.PHONE_NUMBER.value
            else:
                dtype = ProfilerDataTypes.STRING.value
        elif is_object_value(value):
            dtype = ProfilerDataTypes.OBJECT.value

        return dtype

    def levenshtein(self, col_A, col_B):
        return jellyfish.levenshtein_distance(col_A,col_B)


import dask
from dask.distributed import Client, get_client

from optimus.engines.dask.create import Create
from optimus.engines.base.engine import BaseEngine
from optimus.engines.base.remote import MAX_TIMEOUT, RemoteOptimusInterface, RemoteDummyVariable, RemoteDummyDataFrame
from optimus.engines.dask.dask import Dask
from optimus.engines.dask.dataframe import DaskDataFrame
from optimus.engines.dask.io.load import Load
from optimus.optimus import Engine
from optimus._version import __version__


class DaskEngine(BaseEngine):
    __version__ = __version__

    # Using procces or threads https://stackoverflow.com/questions/51099685/best-practices-in-setting-number-of-dask-workers
    def __init__(self, session=None, address=None, n_workers=1, threads_per_worker=None, processes=False,
                 memory_limit='4GB', verbose=False, coiled_token=None, *args, **kwargs):

        if n_workers is None:
            import psutil
            threads_per_worker = psutil.cpu_count() * 4

        self.verbose(verbose)

        use_remote = kwargs.get("use_remote", coiled_token is not None)
        
        if kwargs.get("use_remote", None):
            del kwargs["use_remote"]

        if coiled_token:
            import coiled
            dask.config.set({"coiled.token": coiled_token})
            try:
                coiled.Cloud()
            except Exception as error:
                raise error

            idle_timeout = kwargs.get("idle_timeout", None)

            cluster = coiled.Cluster(
                name=kwargs.get("name"),
                worker_options={
                    **({"nthreads": threads_per_worker} if threads_per_worker else {}),
                    **({"memory_limit": memory_limit} if memory_limit else {})
                },
                n_workers=n_workers,
                worker_memory='15GiB',
                scheduler_options={
                    **({"idle_timeout": idle_timeout} if idle_timeout else {})
                },
                software="optimus/default"
            )

            self.cluster_name = cluster.name
            self.client = Client(cluster)


        elif address:
            self.client = Client(address=address)

        elif session and session != "local":
            self.client = Dask().load(session)

        else:
            try:
                self.client = get_client()
            except ValueError:
                self.client = Client(address=address, n_workers=n_workers, threads_per_worker=threads_per_worker,
                                     processes=processes, memory_limit=memory_limit, *args, **kwargs)

        if use_remote:
            self.remote = RemoteOptimusInterface(self.client, Engine.DASK.value)

        else:
            self.remote = False

    @property
    def dask(self):
        """
        Return a Dask client object
        :return:
        """
        return self.client.dask

    @property
    def constants(self):
        from optimus.engines.base.dask.constants import Constants
        return Constants()

    @property
    def create(self):
        if self.remote:
            return RemoteDummyVariable(self, "_create")

        else:
            return Create(self)

    @property
    def load(self):
        if self.remote:
            return RemoteDummyVariable(self, "_load")

        else:
            return Load(self)

    @property
    def engine(self):
        return Engine.DASK.value

    def dataframe(self, pdf, n_partitions=1, *args, **kwargs):
        from dask import dataframe as dd
        return DaskDataFrame(dd.from_pandas(pdf, npartitions=n_partitions, *args, **kwargs))

    def remote_run(self, func, *args, **kwargs):
        if kwargs.get("client_timeout"):
            client_timeout = kwargs.get("client_timeout")
            del kwargs["client_timeout"]
        else:
            client_timeout = MAX_TIMEOUT

        return self.remote_submit(func, *args, **kwargs).result(client_timeout)

    def remote_submit(self, func, *args, **kwargs):


        if not self.remote:
            fut = self.submit(func, op=self, *args, **kwargs)
        else:
            fut = self.remote.submit(func, *args, **kwargs)

        fut.__result = fut.result
        _op = self

        def _result(self, *args, **kwargs):
            result = self.__result(*args, **kwargs)
            if isinstance(result, dict):
                if result.get("status") == "error" and result.get("error"):
                    raise Exception(result.get("error"))
                elif result.get("dummy"):
                    if result.get("dataframe"):
                        return RemoteDummyDataFrame(_op, result.get("dummy"))
                    else:
                        return RemoteDummyVariable(_op, result.get("dummy"))
            return result

        import types
        fut.result = types.MethodType(_result, fut)

        # TODO: Implement this on remote.submit

        return fut

    def submit(self, func, *args, priority=0, pure=False, **kwargs):
        from optimus.engines.base.remote import RemoteDummyAttribute
        if isinstance(func, (RemoteDummyAttribute,)):
            return func(client_submit=True, *args, **kwargs)
        return dask.distributed.get_client().submit(func, priority=priority, pure=pure, *args, **kwargs)



from optimus.engines.base.dask.dataframe import DaskBaseDataFrame
from optimus.engines.cudf.dataframe import CUDFDataFrame
from optimus.engines.dask.io.save import Save
from optimus.engines.pandas.dataframe import PandasDataFrame
from optimus.helpers.converter import pandas_to_dask_dataframe


class DaskDataFrame(DaskBaseDataFrame):
    def __init__(self, data):
        super().__init__(data)

    def _base_to_dfd(self, pdf, n_partitions):
        return pandas_to_dask_dataframe(pdf, n_partitions)

    @staticmethod
    def pivot(index, column, values):
        pass

    @staticmethod
    def melt(id_vars, value_vars, var_name="variable", value_name="value", data_type="str"):
        pass

    @staticmethod
    def query(sql_expression):
        pass

    @staticmethod
    def debug():
        pass

    @staticmethod
    def create_id(column="id"):
        pass

    @property
    def rows(self):
        from optimus.engines.dask.rows import Rows
        return Rows(self)

    @property
    def cols(self):
        from optimus.engines.dask.columns import Cols
        return Cols(self)

    @property
    def save(self):
        return Save(self)

    @property
    def functions(self):
        from optimus.engines.dask.functions import DaskFunctions
        return DaskFunctions()

    @property
    def mask(self):
        from optimus.engines.dask.mask import DaskMask
        return DaskMask(self)

    @property
    def ml(self):
        from optimus.engines.dask.ml.models import ML
        return ML(self)

    @property
    def encoding(self):
        from optimus.engines.pandas.ml.encoding import Encoding
        return Encoding()

    def to_optimus_pandas(self):
        return PandasDataFrame(self.root.to_pandas())

    def to_optimus_cudf(self):
        return CUDFDataFrame(self.root.to_pandas())



from optimus import Optimus
from pyspark.sql.types import *

op = Optimus()

source_df = op.create.df(
    [('names', StringType(), True), ('height', StringType(), True), ('function', StringType(), True),
     ('rank', StringType(), True)], [('bumbl#ebéé  ', '17.5', 'Espionage', '7'), ("Optim'us", '28.0', 'Leader', '10'),
                                     ('ironhide&', '26.0', 'Security', '7'), ('Jazz', '13.0', 'First Lieutenant', '8'),
                                     ('Megatron', None, 'None', None)])


class TestDataFrameExtensions(object):
    @staticmethod
    def test_melt():
        actual_df = source_df.melt(id_vars=['names'], value_vars=['height', 'function', 'rank'])

        expected_df = op.create.df(
            [('names', StringType(), True), ('variable', StringType(), False), ('value', StringType(), True)],
            [('bumbl#ebéé  ', 'height', '17.5'), ('bumbl#ebéé  ', 'function', 'Espionage'),
             ('bumbl#ebéé  ', 'rank', '7'), ("Optim'us", 'height', '28.0'), ("Optim'us", 'function', 'Leader'),
             ("Optim'us", 'rank', '10'), ('ironhide&', 'height', '26.0'), ('ironhide&', 'function', 'Security'),
             ('ironhide&', 'rank', '7'), ('Jazz', 'height', '13.0'), ('Jazz', 'function', 'First Lieutenant'),
             ('Jazz', 'rank', '8'), ('Megatron', 'height', None), ('Megatron', 'function', 'None'),
             ('Megatron', 'rank', None)])

        assert (expected_df.collect() == actual_df.collect())



import io
import ntpath
import zipfile

import pandas as pd
from packaging import version

from optimus.engines.base.io.load import BaseLoad
from optimus.engines.spark.spark import Spark
from optimus.helpers.columns import replace_columns_special_characters
from optimus.helpers.functions import prepare_path
from optimus.helpers.logger import logger
from optimus.engines.spark.dataframe import SparkDataFrame
from optimus.engines.pandas.dataframe import PandasDataFrame

from optimus.engines.base.meta import Meta


class Load(BaseLoad):

    @staticmethod
    def json(path, multiline=False, *args, **kwargs):
        """
        Return a spark from a json file.
        :param path: path or location of the file.
        :param multiline:
        :return:
        """
        file, file_name = prepare_path(path, "json")

        try:
            df = Spark.instance.spark.read \
                .option("multiLine", multiline) \
                .option("mode", "PERMISSIVE") \
                .json(file, *args, **kwargs)

            df.meta = Meta.set(df.meta, "file_name", file_name)

        except IOError as error:
            print(error)
            raise
        df = replace_columns_special_characters(df)

        df.meta = Meta.set(df.meta, value=df.meta.add_action("columns", df.cols.names()).get())
        return df

    @staticmethod
    def tsv(path, header=True, infer_schema=True, charset="UTF-8", *args, **kwargs):
        """
        Return a spark from a tsv file.
        :param path: path or location of the file.
        :param header: tell the function whether dataset has a header row. True default.
        :param infer_schema: infers the input schema automatically from data.
        :param charset: Charset file encoding
        It requires one extra pass over the data. True default.

        :return:
        """
        df = Load.csv(path, sep='\t', header=header, infer_schema=infer_schema, charset=charset, *args, **kwargs)
        return df

    @staticmethod
    def csv(path, sep=',', header=True, infer_schema=True, encoding="UTF-8", null_value="None", n_rows=-1,
            error_bad_lines=False, *args, **kwargs):
        """
        Return a dataframe from a csv file. It is the same read.csv Spark function with some predefined
        params

        :param path: path or location of the file.
        :param sep: usually delimiter mark are ',' or ';'.
        :param header: tell the function whether dataset has a header row. True default.
        :param infer_schema: infers the input schema automatically from data.
        :param encoding: Charset file encoding
        :param error_bad_lines:
        :param null_value: value to convert the string to a None value
        :param n_rows:
        It requires one extra pass over the data. True default.

        :return dataFrame
        """
        _path, file_name = prepare_path(path, "csv")[0]

        # TODO: Add support to S3 https://bartek-blog.github.io/python/spark/2019/04/22/how-to-access-s3-from-pyspark.html

        try:
            read = (Spark.instance.spark.read
                    .options(header='true' if header else 'false')
                    .options(delimiter=sep)
                    .options(inferSchema='true' if infer_schema else 'false')
                    .options(nullValue=null_value)
                    # .options(quote=null_value)
                    # .options(escape=escapechar)
                    .option("charset", encoding))

            if error_bad_lines is True:
                read.options(mode="FAILFAST")
            else:
                read.options(mode="DROPMALFORMED")

            sdf = read.csv(_path)

            if n_rows > -1:
                sdf = sdf.limit(n_rows)
            # print(type(sdf))
            df = SparkDataFrame(sdf)
            df.meta = Meta.set(df.meta, "file_name", file_name)
        except IOError as error:
            print(error)
            raise

        # df = replace_columns_special_characters(df)

        return df

    @staticmethod
    def parquet(path, *args, **kwargs):
        """
        Return a spark from a parquet file.
        :param path: path or location of the file. Must be string dataType
        :param args: custom argument to be passed to the spark parquet function
        :param kwargs: custom keyword arguments to be passed to the spark parquet function
        :return: Spark Dataframe
        """

        file, file_name = prepare_path(path, "parquet")

        try:
            df = Spark.instance.spark.read.parquet(file, *args, **kwargs)
            df.meta = Meta.set(df.meta, "file_name", file_name)

        except IOError as error:
            print(error)
            raise
        return df

    @staticmethod
    def avro(path, *args, **kwargs):
        """
        Return a spark from a avro file.
        :param path: path or location of the file. Must be string dataType
        :param args: custom argument to be passed to the spark avro function
        :param kwargs: custom keyword arguments to be passed to the spark avro function
        :return: Spark Dataframe
        """
        file, file_name = prepare_path(path, "avro")

        try:
            if version.parse(Spark.instance.spark.version) < version.parse("2.4"):
                avro_version = "com.databricks.spark.avro"
            else:
                avro_version = "avro "
            df = Spark.instance.spark.read.format(avro_version).load(file, *args, **kwargs)

            df.meta = Meta.set(df.meta, "file_name", file_name)
        except IOError as error:
            print(error)
            raise
        return df

    @staticmethod
    def excel(path, sheet_name=0, *args, **kwargs):
        """
        Return a spark from a excel file.
        :param path: Path or location of the file. Must be string dataType
        :param sheet_name: excel sheet name
        :param args: custom argument to be passed to the excel function
        :param kwargs: custom keyword arguments to be passed to the excel function
        :return: Spark Dataframe
        """
        file, file_name = prepare_path(path, "xls")

        try:
            pdf = pd.read_excel(file, sheet_name=sheet_name, *args, **kwargs)

            # Parse object column data type to string to ensure that Spark can handle it. With this we try to reduce
            # exception when Spark try to infer the column data type
            col_names = list(pdf.select_dtypes(include=['object']))

            column_dtype = {}
            for col in col_names:
                column_dtype[col] = str

            # Convert object columns to string
            pdf = pdf.astype(column_dtype)

            # Create spark data frame
            df = Spark.instance.spark.createDataFrame(pdf)
            df.meta = Meta.set(df.meta, "file_name", ntpath.basename(file_name))
        except IOError as error:
            print(error)
            raise

        df = replace_columns_special_characters(df)
        return df

    @staticmethod
    def zip(path, file_name=None):
        """
        Return a Dataframe from a file inside a zip
        :param path:
        :param file:
        :return:
        """

        zip_file, zip_filename = prepare_path(path, "zip")

        def zip_extract(x):
            in_memory_data = io.BytesIO(x[1])
            file_obj = zipfile.ZipFile(in_memory_data, "r")
            files = [i for i in file_obj.namelist()]
            return dict(zip(files, [file_obj.open(file).read() for file in files]))

        zips = Spark.instance.sc.binaryFiles(zip_file)

        files_data = zips.map(zip_extract).collect()
        if file_name is None:
            result = files_data
        else:
            result = files_data[file_name]
        return result




from pyspark.sql.types import *
from optimus import Optimus
from pyspark.sql import functions as F


class TestOptimus(object):
    @staticmethod
    def test_optimus_from_session():
        from pyspark.sql import SparkSession
        spark = SparkSession.builder.appName('abc').getOrCreate()
        df = spark.read.csv('examples/data/foo.csv', header=True)
        op = Optimus(spark)
        df.display()


from pyspark.sql.types import *
from optimus import Optimus
import datetime
from pyspark.sql import functions as F

op = Optimus(master='local')


class TestOptimus(object):
    @staticmethod
    def test_create_df_one_column():
        source_df = op.create.df([('name', StringType(), True)], [('Argenis',), ('Favio',), ('Matthew',)])
        actual_df = source_df
        expected_df = op.create.df([('name', StringType(), True)], [('Argenis',), ('Favio',), ('Matthew',)])
        assert (expected_df.collect() == actual_df.collect())

    @staticmethod
    def test_create_df_plain():
        source_df = op.create.df([('name', StringType(), True), ('age', IntegerType(), True)],
                                 [('BOB', 1), ('JoSe', 2)])
        actual_df = source_df
        expected_df = op.create.df([('name', StringType(), True), ('age', IntegerType(), True)],
                                   [('BOB', 1), ('JoSe', 2)])
        assert (expected_df.collect() == actual_df.collect())

    @staticmethod
    def test_create_df_plain_infer_false():
        source_df = op.create.df([('name', StringType(), True), ('age', StringType(), True)],
                                 [('BOB', '1'), ('JoSe', '2')])
        actual_df = source_df
        expected_df = op.create.df([('name', StringType(), True), ('age', StringType(), True)],
                                   [('BOB', '1'), ('JoSe', '2')])
        assert (expected_df.collect() == actual_df.collect())

    @staticmethod
    def test_create_df_with_data_types():
        source_df = op.create.df([('name', StringType(), True), ('age', IntegerType(), False)],
                                 [('BOB', 1), ('JoSe', 2)])
        actual_df = source_df
        expected_df = op.create.df([('name', StringType(), True), ('age', IntegerType(), False)],
                                   [('BOB', 1), ('JoSe', 2)])
        assert (expected_df.collect() == actual_df.collect())

    @staticmethod
    def test_create_df_nullable():
        source_df = op.create.df([('name', StringType(), True), ('age', IntegerType(), True)],
                                 [('BOB', 1), ('JoSe', 2)])
        actual_df = source_df
        expected_df = op.create.df([('name', StringType(), True), ('age', IntegerType(), True)],
                                   [('BOB', 1), ('JoSe', 2)])
        assert (expected_df.collect() == actual_df.collect())



# from dask_ml.preprocessing import DummyEncoder

from optimus.helpers.check import is_spark_dataframe
from optimus.helpers.columns import parse_columns, name_col
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_, is_str


def n_gram(df, input_col, n=2):
    """
    Converts the input array of strings inside of a dataframe into an array of n-grams.
    :param df: Dataframe to analyze
    :param input_col: Column to analyzer.
    :param n: number of elements per n-gram >=1.
    :return: Spark DataFrame with n-grams calculated.
    """

    is_spark_dataframe(df)

    tokenizer = feature.Tokenizer().setInputCol(input_col) | feature.StopWordsRemover()
    count = feature.CountVectorizer()
    gram = feature.NGram(n=n) | feature.CountVectorizer()
    tf = tokenizer | (count, gram) | feature.VectorAssembler()
    tfidf = tf | feature.IDF().setOutputCol('features')

    tfidf_model = tfidf.fit(df)
    df_model = tfidf_model.transform(df)
    return df_model, tfidf_model


def one_hot_encoder(df, input_cols, output_col=None, **kargs):
    """
    Maps a column of label indices to a column of binary vectors, with at most a single one-value.
    :param df: Dataframe to be transformed.
    :param input_cols: Columns to be encoded.
    :param output_col: Column where the output is going to be saved.
    :return: Dataframe with encoded columns.
    """

    input_cols = parse_columns(df, input_cols)

    if output_col is None:
        output_col = name_col(input_cols, "one_hot_encoder")

    de = DummyEncoder()
    df[output_col] = de.fit_transform(df[input_cols])

    return df


# TODO: Must we use the pipeline version?
def vector_assembler(df, input_cols, output_col=None):
    """
    Combines a given list of columns into a single vector column.
    :param df: Dataframe to be transformed.
    :param input_cols: Columns to be assembled.
    :param output_col: Column where the output is going to be saved.
    :return: Dataframe with assembled column.
    """

    input_cols = parse_columns(df, input_cols)

    if output_col is None:
        output_col = name_col(input_cols, "vector_assembler")

    assembler = [VectorAssembler(inputCols=input_cols, outputCol=output_col)]

    pipeline = Pipeline(stages=assembler)
    df = pipeline.fit(df).transform(df)

    return df


def normalizer(df, input_cols, output_col=None, p=2.0):
    """
    Transforms a dataset of Vector rows, normalizing each Vector to have unit norm. It takes parameter p, which
    specifies the p-norm used for normalization. (p=2) by default.
    :param df: Dataframe to be transformed
    :param input_cols: Columns to be normalized.
    :param output_col: Column where the output is going to be saved.
    :param p:  p-norm used for normalization.
    :return: Dataframe with normalized columns.
    """

    # Check if columns argument must be a string or list datat ype:
    if not is_(input_cols, (str, list)):
        RaiseIt.type_error(input_cols, ["str", "list"])

    if is_str(input_cols):
        input_cols = [input_cols]

    if is_(input_cols, (float, int)):
        RaiseIt.type_error(input_cols, ["float", "int"])

    # Try to create a vector
    if len(input_cols) > 1:
        df = df.cols.cast(input_cols, "vector")

    if output_col is None:
        output_col = name_col(input_cols, "normalizer")

    # TODO https://developer.ibm.com/code/2018/04/10/improve-performance-ml-pipelines-wide-dataframes-apache-spark-2-3/
    normal = [Normalizer(inputCol=col_name, outputCol=output_col, p=p) for col_name in
              list(set(input_cols))]

    pipeline = Pipeline(stages=normal)

    df = pipeline.fit(df).transform(df)

    return df



from pandas import DataFrame as PandasDataFrame

from optimus.engines.base.basedataframe import BaseDataFrame

MAX_TIMEOUT = 600


class RemoteDummyAttribute:

    def __init__(self, name, names, dummy_id, op):
        self.__names = [*names, name]
        self.__op = op
        self.__id = dummy_id

    def __getattr__(self, item):
        return RemoteDummyAttribute(item, self.__names, self.__id, self.__op)

    def __call__(self, *args, **kwargs):

        if kwargs.get("client_submit"):
            client_submit = kwargs["client_submit"]
            del kwargs["client_submit"]
        else:
            client_submit = False

        def _f(op, unique_id, method, *args, **kwargs):
            obj = op.get_var(unique_id)
            if obj is None:
                op.del_var(unique_id)
                raise Exception("Remote variable with id " + unique_id + " not found or null")
            func = obj
            for me in method:
                func = getattr(func, me)
            if callable(func):
                result = func(*args, **kwargs)
            else:
                result = func
            return result

        if client_submit:
            return self.__op.remote_submit(_f, self.__id, self.__names, *args, **kwargs)
        else:
            return self.__op.remote_run(_f, self.__id, self.__names, *args, **kwargs)


class RemoteDummyVariable:

    def __init__(self, op, unique_id, *args, **kwargs):
        self.op = op
        self.id = unique_id

    def __getattr__(self, item):
        if item.startswith('_'):
            raise AttributeError(item)
        return RemoteDummyAttribute(item, [], self.id, self.op)

    def __getstate__(self):
        return {"op": self.op, "id": self.id}

    def __setstate__(self, d):
        self.op = d.op
        self.id = d.id
        return

    def __del__(self):
        self.op.remote.del_var(self.id).result(180)


class RemoteDummyDataFrame(RemoteDummyVariable):
    print = BaseDataFrame.print
    table = BaseDataFrame.table
    display = BaseDataFrame.display

    def __repr__(self):
        return self.ascii()

    def _repr_html_(self):
        return self.table()

    def __len__(self):
        return self.rows.count()

    @property
    def meta(self):
        def _get_attr(op, unique_id, attr):
            df = op.get_var(unique_id)
            if df is None:
                op.del_var(unique_id)
                raise Exception("Remote variable with id " + unique_id + " not found or null")
            return getattr(df, attr)

        return self.op.remote_run(_get_attr, self.id, "meta")


class RemoteOptimusInterface:
    op = {}
    _vars = {}
    _del_next = []

    def __init__(self, client, engine=False):
        if not engine:
            from optimus.optimus import Engine
            engine = Engine.DASK.value

        self.client = client
        self.engine = engine

        future = self._create_actor(self.engine)
        future.result(MAX_TIMEOUT)

    def _create_actor(self, engine):
        self.engine = engine

        def _init(_engine):
            from dask.distributed import get_worker
            worker = get_worker()
            worker.actor = RemoteOptimus(_engine)
            return f"Created remote Optimus instance using \"{_engine}\""

        return self.client.submit(_init, self.engine)

    def submit(self, func, *args, priority=0, pure=False, **kwargs):
        def _remote(_func, *args, **kwargs):
            from dask.distributed import get_worker
            actor = get_worker().actor
            return actor.submit(_func, priority=priority, pure=pure, *args, **kwargs)

        kwargs.update({"pure": False})

        return self.client.submit(_remote, func, *args, **kwargs)

    def run(self, func, *args, **kwargs):
        if kwargs.get("client_timeout"):
            client_timeout = kwargs.get("client_timeout")
            del kwargs["client_timeout"]
        else:
            client_timeout = MAX_TIMEOUT

        return self.submit(func, *args, **kwargs).result(client_timeout)

    def list_vars(self, client_timeout=MAX_TIMEOUT):
        def _list_vars():
            from dask.distributed import get_worker
            op = get_worker().actor.op
            return op.list_vars()

        return self.client.submit(_list_vars, pure=False).result(client_timeout)

    def clear_vars(self, keep=[], client_timeout=MAX_TIMEOUT):
        def _clear_vars(keep=[]):
            from dask.distributed import get_worker
            op = get_worker().actor.op
            return op.clear_vars(keep)

        return self.client.submit(_clear_vars, keep, pure=False).result(client_timeout)

    def update_vars(self, values, client_timeout=MAX_TIMEOUT):
        def _update_vars(values):
            from dask.distributed import get_worker
            op = get_worker().actor.op
            return op.update_vars(values)

        return self.client.submit(_update_vars, values, pure=False).result(client_timeout)

    def del_var(self, name, client_timeout=MAX_TIMEOUT):
        def _del_var(name):
            from dask.distributed import get_worker
            op = get_worker().actor.op
            return op.del_var(name)

        return self.client.submit(_del_var, name, pure=False).result(client_timeout)

    def set_var(self, name, value, client_timeout=MAX_TIMEOUT):
        def _set_var(name):
            from dask.distributed import get_worker
            op = get_worker().actor.op
            return op.set_var(name, value)

        return self.client.submit(_set_var, name, value, pure=False).result(client_timeout)

    def get_var(self, name, client_timeout=MAX_TIMEOUT):
        def _get_var(name):
            from dask.distributed import get_worker
            op = get_worker().actor.op
            return op.get_var(name)

        return self.client.submit(_get_var, name, pure=False).result(client_timeout)


class RemoteOptimus:
    op = {}
    _vars = {}
    _del_next = []

    def __init__(self, engine=False):
        from optimus.optimus import Engine

        if not engine:
            engine = Engine.DASK.value

        from optimus import Optimus
        self.op = Optimus(engine)
        self.op.set_var = self.set_var
        self.op.get_var = self.get_var
        self.op.del_var = self.del_var
        self.op.clear_vars = self.clear_vars
        self.op.list_vars = self.list_vars
        self.op.update_vars = self.update_vars
        self.set_var("_load", self.op.load)
        self.set_var("_create", self.op.create)

        import numpy as np
        if engine == Engine.DASK_CUDF.value:
            import cupy as cp
            self.allowed_types = (str, bool, int, float, complex, np.generic, cp.generic)
        else:
            self.allowed_types = (str, bool, int, float, complex, np.generic)

    def list_vars(self):
        return list(self._vars.keys())

    def clear_vars(self, keep=[]):
        keep = keep + ["_load", "_create"]
        self._vars = {k: self._vars[k] for k in keep}
        return list(self._vars.keys())

    def update_vars(self, values):
        self._vars.update(values)

    def _del_var(self, name):
        try:
            del self._vars[name]
        except:
            print(name + " not found")

    def del_var(self, name):

        for _name in self._del_next:
            self._del_var(_name)

        self._del_next = []

        if not name.startswith("_"):
            if name not in self._vars or self._vars[name] is None:
                print(name + " not found")
            else:
                self._del_next.append(name)

    def set_var(self, name, value):
        self._vars[name] = value

    def get_var(self, name):
        return self._vars.get(name, None)

    def _return(self, value):
        if isinstance(value, (dict,)):
            for key in value:
                value[key] = self._return(value[key])
            return value
        elif isinstance(value, (list,)):
            return list(map(self._return, value))
        elif isinstance(value, (set,)):
            return set(map(self._return, value))
        elif isinstance(value, (tuple,)):
            return tuple(map(self._return, value))
        elif isinstance(value, (PandasDataFrame,)):
            return value.head()
        elif not isinstance(value, self.allowed_types) and value is not None:
            import uuid
            unique_id = str(uuid.uuid4())
            self.set_var(unique_id, value)
            if isinstance(value, (BaseDataFrame,)):
                return {"dummy": unique_id, "dataframe": True}
            else:
                return {"dummy": unique_id, "dataframe": False}
        else:
            return value

    def submit(self, callback, *args, **kwargs):
        try:
            result = callback(self.op, *args, **kwargs)
        except Exception as err:
            import traceback
            error_class = err.__class__.__name__
            detail = err.args[0]
            tb = traceback.format_exc()
            error = "%s: %s\n%s" % (error_class, detail, tb)
            return {"status": "error", "error": error}
        return self._return(result) if result is not None else None



from optimus.engines.base.create import BaseCreate
from optimus.helpers.types import DataFrameType

from optimus.engines.spark.dataframe import SparkDataFrame
from optimus.engines.spark.spark import Spark


class Create(BaseCreate):

    def _df_from_dfd(self, dfd, *args, **kwargs) -> DataFrameType:
        dfd = Spark.instance.spark.createDataFrame(dfd)
        return SparkDataFrame(dfd, *args, **kwargs)



from enum import Enum

from optimus.engines.base.io.drivers.abstract_driver import AbstractDriver
from optimus.engines.spark.io.properties import DriverProperties
from optimus.helpers.functions import singleton


@singleton
class SQLiteDriver(AbstractDriver):
    """SQLite Database"""

    def properties(self) -> Enum:
        return DriverProperties.SQLITE

    def url(self, *args, **kwargs) -> str:
        return f"""jdbc:{kwargs["driver"]}:{kwargs["host"]}"""

    def table_names_query(self, *args, **kwargs) -> str:
        return "SELECT name FROM sqlite_master WHERE type='table'"

    def table_name_query(self, *args, **kwargs) -> str:
        return "SELECT name FROM sqlite_master WHERE type='table'"

    def count_query(self, *args, **kwargs) -> str:
        return "SELECT COUNT(*) as COUNT FROM " + kwargs["db_table"]

    def primary_key_query(self, *args, **kwargs) -> str:
        pass

    def min_max_query(self, *args, **kwargs) -> str:
        return f"""SELECT min({kwargs["partition_column"]}) AS min, max({kwargs["partition_column"]}) AS max FROM {
        kwargs["table_name"]} """



from abc import ABC

from optimus.engines.base.meta import Meta
from optimus.helpers.columns import parse_columns
from optimus.helpers.core import one_list_to_val
from optimus.helpers.json import dump_json
from optimus.infer import is_list
from optimus.profiler.constants import MAX_BUCKETS


class BaseProfile(ABC):
    """Base class for all profile implementations"""

    def __init__(self, root):
        self.root = root

    def summary(self, columns="*"):

        df = self.root

        return Meta.get(df.meta, f"profile.summary")

    def columns(self, columns="*"):

        df = self.root
        columns = parse_columns(df, columns) if columns else []

        if is_list(columns):
            columns = [Meta.get(df.meta, f"profile.columns.{col}") for col in columns]
        else:
            columns = Meta.get(df.meta, f"profile.columns.{columns}")

        return one_list_to_val(columns)

    def dtypes(self, columns="*"):

        df = self.root
        columns = parse_columns(df, columns) if columns else []

        if is_list(columns):
            dtype = [Meta.get(df.meta, f"profile.columns.{col}.stats.profiler_dtype.dtype") for col in columns]
        else:
            dtype = Meta.get(df.meta, f"profile.columns.{columns}.stats.profiler_dtype.dtype")

        return one_list_to_val(dtype)

    def __call__(self, columns="*", bins: int = MAX_BUCKETS, output: str = None, flush: bool = False, size=False) -> dict:
        # def profile(self, columns="*", bins: int = MAX_BUCKETS, output: str = None, flush: bool = False, size=False):
        """
        Return a dict the profile of the dataset
        :param columns:
        :param bins:
        :param output:
        :param flush:
        :param size: get the dataframe size in memory. Use with caution this could be slow for big data frames.
        :return:
        """

        df = self.root

        meta = self.root.meta
        profile = Meta.get(meta, "profile")

        if not profile:
            flush = True

        if columns or flush:
            columns = parse_columns(df, columns) if columns else []
            transformations = Meta.get(meta, "transformations") or []

            calculate = False

            if flush or len(transformations):
                calculate = True

            else:
                for col in columns:
                    if col not in profile["columns"]:
                        calculate = True

            if calculate:
                df = df[columns].calculate_profile(columns, bins, flush, size)
                profile = Meta.get(df.meta, "profile")
                self.root.meta = df.meta
            profile["columns"] = {key: profile["columns"][key] for key in columns}

        if output == "json":
            profile = dump_json(profile)

        return profile



import datetime

import simplejson as json
import pandas as pd
import numpy as np


def json_converter(obj):
    """
    Custom converter to be used with json.dumps
    :param obj:
    :return:
    """

    if not pd.isnull(obj):
        if isinstance(obj, datetime.datetime):
            # return obj.strftime('%Y-%m-%d %H:%M:%S')
            return obj.isoformat()

        elif isinstance(obj, datetime.date):
            return obj.isoformat()
            # return obj.strftime('%Y-%m-%d')

        elif isinstance(obj, (np.generic,)):
            return np.asscalar(obj)


def json_enconding(obj):
    """
    Encode a json. Used for testing.
    :param obj:
    :return:
    """
    return json.dumps(obj, default=json_converter)


def dump_json(value, *args, **kwargs):

    def _replace(data):
        if isinstance(data, dict):
            return {k: _replace(v) for k, v in data.items()}
        elif isinstance(data, list):
            return [_replace(i) for i in data]
        elif data == float("inf"):
            return "Infinity"
        elif data == float("-inf"):
            return "-Infinity"
        else:
            return data
    return json.dumps(_replace(value), ignore_nan=True, default=json_converter, *args, **kwargs)



from enum import Enum

from optimus.engines.base.io.drivers.abstract_driver import AbstractDriver
from optimus.engines.spark.io.properties import DriverProperties
from optimus.helpers.functions import singleton


@singleton
class CassandraDriver(AbstractDriver):
    """Cassandra Database"""

    def properties(self) -> Enum:
        return DriverProperties.CASSANDRA

    def url(self, *args, **kwargs) -> str:
        return ""

    def table_names_query(self, *args, **kwargs) -> str:
        pass

    def table_name_query(self, *args, **kwargs) -> str:
        pass

    def count_query(self, *args, **kwargs) -> str:
        pass

    def primary_key_query(self, *args, **kwargs) -> str:
        pass

    def min_max_query(self, *args, **kwargs) -> str:
        return f"""SELECT min({kwargs["partition_column"]}) AS min, max({kwargs["partition_column"]}) AS max FROM {
        kwargs["table_name"]} """



# import platform
# import papermill as pm
# import numpy as np
#
# nan = np.nan
#
# notebooks = ["10_min_from_spark_to_pandas_with_optimus.ipynb",
#              "column.ipynb",
#              "row.ipynb",
#              "profiler.ipynb",
#              "plot.ipynb.ipynb",
#              "optimus.ipynb",
#              "hacking_optimus.ipynb",
#              "spark-optimus.ipynb",
#              "ml.ipynb",
#              "outliers.ipynb"]
#
#
# class Test_example_notebooks(object):
#     @staticmethod
#     def test_notebooks():
#         p = platform.system()
#         if p == "Linux" or p == "Darwin":
#             output = "/dev/null"
#         elif p == "Windows":
#             output = "$null"
#
#         for notebook in notebooks:
#             pm.execute_notebook(
#                 "/home/travis/build/hi-primus/optimus/examples/" + notebook,
#                 output
#             )



from pyspark.sql.types import *
from optimus import Optimus
from optimus.helpers.json import json_enconding
from optimus.helpers.functions import deep_sort
import unittest
from pyspark.ml.linalg import Vectors, VectorUDT, DenseVector
import numpy as np
nan = np.nan
import datetime
from pyspark.sql import functions as F
from optimus.profiler.profiler import Profiler
null = None
true = True
op = Optimus(master='local')
source_df=op.create.df([('names', StringType(), True),('height(ft)', ShortType(), True),('function', StringType(), True),('rank', ByteType(), True),('age', IntegerType(), True),('weight(t)', FloatType(), True),('japanese name', ArrayType(StringType(),True), True),('last position seen', StringType(), True),('date arrival', StringType(), True),('last date seen', StringType(), True),('attributes', ArrayType(FloatType(),True), True),('Date Type', DateType(), True),('timestamp', TimestampType(), True),('Cybertronian', BooleanType(), True),('function(binary)', BinaryType(), True),('NullType', NullType(), True)], [('Optimus', -28, 'Leader', 10, 5000000, 4.300000190734863, ['Inochi', 'Convoy'], '19.442735,-99.201111', '1980/04/10', '2016/09/10', [8.53439998626709, 4300.0], datetime.date(2016, 9, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Leader'), None), ('bumbl#ebéé  ', 17, 'Espionage', 7, 5000000, 2.0, ['Bumble', 'Goldback'], '10.642707,-71.612534', '1980/04/10', '2015/08/10', [5.334000110626221, 2000.0], datetime.date(2015, 8, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Espionage'), None), ('ironhide&', 26, 'Security', 7, 5000000, 4.0, ['Roadbuster'], '37.789563,-122.400356', '1980/04/10', '2014/07/10', [7.924799919128418, 4000.0], datetime.date(2014, 6, 24), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Security'), None), ('Jazz', 13, 'First Lieutenant', 8, 5000000, 1.7999999523162842, ['Meister'], '33.670666,-117.841553', '1980/04/10', '2013/06/10', [3.962399959564209, 1800.0], datetime.date(2013, 6, 24), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'First Lieutenant'), None), ('Megatron', None, 'None', 10, 5000000, 5.699999809265137, ['Megatron'], None, '1980/04/10', '2012/05/10', [None, 5700.0], datetime.date(2012, 5, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'None'), None), ('Metroplex_)^$', 300, 'Battle Station', 8, 5000000, None, ['Metroflex'], None, '1980/04/10', '2011/04/10', [91.44000244140625, None], datetime.date(2011, 4, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Battle Station'), None)])
class Test_df_ext(unittest.TestCase):
	maxDiff = None
	@staticmethod
	def test_ext_sample():
		actual_df =source_df.sample(5,True)
		expected_df = op.create.df([('names', StringType(), True),('height(ft)', ShortType(), True),('function', StringType(), True),('rank', ByteType(), True),('age', IntegerType(), True),('weight(t)', FloatType(), True),('japanese name', ArrayType(StringType(),True), True),('last position seen', StringType(), True),('date arrival', StringType(), True),('last date seen', StringType(), True),('attributes', ArrayType(FloatType(),True), True),('Date Type', DateType(), True),('timestamp', TimestampType(), True),('Cybertronian', BooleanType(), True),('function(binary)', BinaryType(), True),('NullType', NullType(), True)], [('Optimus', -28, 'Leader', 10, 5000000, 4.300000190734863, ['Inochi', 'Convoy'], '19.442735,-99.201111', '1980/04/10', '2016/09/10', [8.53439998626709, 4300.0], datetime.date(2016, 9, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Leader'), None), ('bumbl#ebéé  ', 17, 'Espionage', 7, 5000000, 2.0, ['Bumble', 'Goldback'], '10.642707,-71.612534', '1980/04/10', '2015/08/10', [5.334000110626221, 2000.0], datetime.date(2015, 8, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Espionage'), None), ('ironhide&', 26, 'Security', 7, 5000000, 4.0, ['Roadbuster'], '37.789563,-122.400356', '1980/04/10', '2014/07/10', [7.924799919128418, 4000.0], datetime.date(2014, 6, 24), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Security'), None), ('Megatron', None, 'None', 10, 5000000, 5.699999809265137, ['Megatron'], None, '1980/04/10', '2012/05/10', [None, 5700.0], datetime.date(2012, 5, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'None'), None), ('Metroplex_)^$', 300, 'Battle Station', 8, 5000000, None, ['Metroflex'], None, '1980/04/10', '2011/04/10', [91.44000244140625, None], datetime.date(2011, 4, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Battle Station'), None)])
		assert (expected_df.collect() == actual_df.collect())
	@staticmethod
	def test_ext_stratified_sample():
		actual_df =source_df.stratified_sample('rank',12)
		expected_df = op.create.df([('names', StringType(), True),('height(ft)', ShortType(), True),('function', StringType(), True),('rank', ByteType(), True),('age', IntegerType(), True),('weight(t)', FloatType(), True),('japanese name', ArrayType(StringType(),True), True),('last position seen', StringType(), True),('date arrival', StringType(), True),('last date seen', StringType(), True),('attributes', ArrayType(FloatType(),True), True),('Date Type', DateType(), True),('timestamp', TimestampType(), True),('Cybertronian', BooleanType(), True),('function(binary)', BinaryType(), True),('NullType', NullType(), True)], [('Optimus', -28, 'Leader', 10, 5000000, 4.300000190734863, ['Inochi', 'Convoy'], '19.442735,-99.201111', '1980/04/10', '2016/09/10', [8.53439998626709, 4300.0], datetime.date(2016, 9, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Leader'), None), ('bumbl#ebéé  ', 17, 'Espionage', 7, 5000000, 2.0, ['Bumble', 'Goldback'], '10.642707,-71.612534', '1980/04/10', '2015/08/10', [5.334000110626221, 2000.0], datetime.date(2015, 8, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Espionage'), None), ('ironhide&', 26, 'Security', 7, 5000000, 4.0, ['Roadbuster'], '37.789563,-122.400356', '1980/04/10', '2014/07/10', [7.924799919128418, 4000.0], datetime.date(2014, 6, 24), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Security'), None), ('Jazz', 13, 'First Lieutenant', 8, 5000000, 1.7999999523162842, ['Meister'], '33.670666,-117.841553', '1980/04/10', '2013/06/10', [3.962399959564209, 1800.0], datetime.date(2013, 6, 24), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'First Lieutenant'), None), ('Megatron', None, 'None', 10, 5000000, 5.699999809265137, ['Megatron'], None, '1980/04/10', '2012/05/10', [None, 5700.0], datetime.date(2012, 5, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'None'), None)])
		assert (expected_df.collect() == actual_df.collect())



import functools
import operator

import dask.array as da
import dask.dataframe as dd
from dask.delayed import delayed
from multipledispatch import dispatch

from optimus.engines.base.meta import Meta
from optimus.engines.base.rows import BaseRows
from optimus.helpers.columns import parse_columns
from optimus.helpers.constants import Actions
from optimus.helpers.core import val_to_list, one_list_to_val
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_list_of_str_or_int, is_list_value


class DaskBaseRows(BaseRows):
    """Base class for all Rows implementations"""

    def __init__(self, parent):
        # self.parent = parent
        super().__init__(parent)
        # super(DaskBaseRows, self).__init__(parent)

    def _reverse(self, dfd):
        @delayed
        def reverse_pdf(pdf):
            return pdf[::-1]

        ds = dfd.to_delayed()
        ds = [reverse_pdf(d) for d in ds][::-1]
        return dd.from_delayed(ds)

    def _sort(self, dfd, col_name, ascending):
        dfd = dfd.set_index(col_name)
        if not ascending:
            dfd = self._reverse(dfd)
        return dfd.reset_index()[self.root.cols.names()]

    def create_id(self, column="id"):
        # Reference https://github.com/dask/dask/issues/1426
        dfd = self.root.data
        # print(dfd)
        a = da.arange(dfd.divisions[-1] + 1, chunks=dfd.divisions[1:])
        dfd[column] = dd.from_dask_array(a)
        return dfd

    def append(self, dfs, names_map=None):
        """
        Appends 2 or more dataframes
        :param dfs:
        :param names_map:
        """
        if not is_list_value(dfs):
            dfs = [dfs]

        every_df = [self.root, *dfs]

        if names_map is not None:
            rename = [[] for _ in every_df]
            for key in names_map:
                assert len(names_map[key]) == len(every_df)
                for i in range(len(names_map[key])):
                    col_name = names_map[key][i]
                    if col_name:
                        rename[i] = [*rename[i], (col_name, "__output_column__" + key)]
            for i in range(len(rename)):
                every_df[i] = every_df[i].cols.rename(rename[i])

        dfd = every_df[0].data
        for i in range(len(every_df)):
            if i != 0:
                dfd = dfd.append(every_df[i].data)
        df = self.root.new(dfd)

        if names_map is not None:
            df = df.cols.rename([("__output_column__" + key, key) for key in names_map])
            df = df.cols.select([*names_map.keys()])

        return df.new(df.data.reset_index(drop=True))

    # def append(self, rows):
    #     """
    #
    #     :param rows:
    #     :return:
    #     """
    #     dfd = self.root.data
    #
    #     if is_list(rows):
    #         rows = dd.from_pandas(pd.DataFrame(rows), npartitions=1)
    #
    #     # Can not concatenate dataframe with not string columns names
    #     rows.columns = dfd.columns
    #
    #     dfd = dd.concat([dfd, rows], axis=0, interleave_partitions=True)
    #
    #     return dfd

    def limit(self, count):
        """
        Limit the number of rows
        :param count:
        :return:
        """
        df = self.root
        # Reference https://stackoverflow.com/questions/49139371/slicing-out-a-few-rows-from-a-dask-dataframe

        if count is None:
            return df

        partitions = df.partitions()

        return self.root.new(self.root._base_to_dfd(df.cols.select("*").data.head(count), partitions))

    def between(self, columns, lower_bound=None, upper_bound=None, invert=False, equal=False,
                bounds=None):
        """
        Trim values at input thresholds
        :param upper_bound:
        :param lower_bound:
        :param columns: Columns to be trimmed
        :param invert:
        :param equal:
        :param bounds:
        :return:
        """
        df = self.root
        # TODO: should process string or dates
        # columns = parse_columns(df, columns, filter_by_column_dtypes=df.constants.NUMERIC_TYPES)
        columns = parse_columns(df, columns)
        if bounds is None:
            bounds = [(lower_bound, upper_bound)]

        def _between(_col_name):

            if invert is False and equal is False:
                op1 = operator.gt
                op2 = operator.lt
                opb = operator.__and__

            elif invert is False and equal is True:
                op1 = operator.ge
                op2 = operator.le
                opb = operator.__and__

            elif invert is True and equal is False:
                op1 = operator.lt
                op2 = operator.gt
                opb = operator.__or__

            elif invert is True and equal is True:
                op1 = operator.le
                op2 = operator.ge
                opb = operator.__or__

            sub_query = []
            for bound in bounds:
                _lower_bound, _upper_bound = bound
                sub_query.append(opb(op1(df[_col_name], _lower_bound), op2(df[_col_name], _upper_bound)))
            query = functools.reduce(operator.__or__, sub_query)

            return query

        for col_name in columns:
            df = df.rows.select(_between(col_name))
        meta = Meta.action(df.meta, Actions.DROP_ROW.value, df.cols.names())
        return self.root.new(df.data, meta=meta)

    def unnest(self, input_cols):
        df = self.root
        return df

    def approx_count(self):
        """
        Aprox rows count
        :return:
        """
        df = self.root
        return df.rows.count()



from pyspark.ml import feature, Pipeline
from pyspark.ml.feature import StringIndexer, OneHotEncoder, VectorAssembler, Normalizer, IndexToString

from optimus.helpers.columns import parse_columns, name_col, get_output_cols
from optimus.helpers.constants import Actions
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_, is_str
from optimus.helpers.check import is_spark_dataframe
from optimus.engines.base.ml.contants import STRING_TO_INDEX, INDEX_TO_STRING

from optimus.engines.base.meta import Meta


def n_gram(df, input_col, n=2):
    """
    Converts the input array of strings inside of a Spark DF into an array of n-grams.
    :param df: Pyspark dataframe to analyze
    :param input_col: Column to analyzer.
    :param n: number of elements per n-gram >=1.
    :return: Spark DataFrame with n-grams calculated.
    """

    is_spark_dataframe(df)

    tokenizer = feature.Tokenizer().setInputCol(input_col) | feature.StopWordsRemover()
    count = feature.CountVectorizer()
    gram = feature.NGram(n=n) | feature.CountVectorizer()
    tf = tokenizer | (count, gram) | feature.VectorAssembler()
    tfidf = tf | feature.IDF().setOutputCol('features')

    tfidf_model = tfidf.fit(df)
    df_model = tfidf_model.transform(df)
    return df_model, tfidf_model


def string_to_index(df, input_cols, output_cols=None, columns=None, **kargs):
    """
    Maps a string column of labels to an ML column of label indices. If the input column is
    numeric, we cast it to string and index the string values.
    :param df: Dataframe to be transformed
    :param input_cols: Columns to be indexed.
    :param output_cols:Column where the ouput is going to be saved
    :param columns:
    :return: Dataframe with indexed columns.
    """
    df_actual = df

    if columns is None:
        input_cols = parse_columns(df, input_cols)
        if output_cols is None:
            output_cols = [name_col(input_col, STRING_TO_INDEX) for input_col in input_cols]
        output_cols = get_output_cols(input_cols, output_cols)
    else:
        input_cols, output_cols = zip(*columns)

    indexers = [StringIndexer(inputCol=input_col, outputCol=output_col, **kargs).fit(df) for input_col, output_col
                in zip(list(set(input_cols)), list(set(output_cols)))]

    pipeline = Pipeline(stages=indexers)
    df = pipeline.fit(df).transform(df)

    df.meta = Meta.action(df.meta, df_actual, Actions.STRING_TO_INDEX.value, output_cols)

    return df


def index_to_string(df, input_cols, output_cols=None, columns=None, **kargs):
    """
    Maps a column of indices back to a new column of corresponding string values. The index-string mapping is
    either from the ML attributes of the input column, or from user-supplied labels (which take precedence over
    ML attributes).
    :param df: Dataframe to be transformed.
    :param input_cols: Columns to be indexed.
    :param output_cols: Column where the output is going to be saved.
    :param columns:
    :return: Dataframe with indexed columns.
    """
    df_actual = df

    if columns is None:
        input_cols = parse_columns(df, input_cols)
        if output_cols is None:
            output_cols = [name_col(input_col, INDEX_TO_STRING) for input_col in input_cols]
        output_cols = get_output_cols(input_cols, output_cols)
    else:
        input_cols, output_cols = zip(*columns)

    indexers = [IndexToString(inputCol=input_col, outputCol=output_col, **kargs) for input_col, output_col
                in zip(list(set(input_cols)), list(set(output_cols)))]
    pipeline = Pipeline(stages=indexers)
    df = pipeline.fit(df).transform(df)

    df.meta = Meta.action(df.meta, df_actual, Actions.INDEX_TO_STRING.value, output_cols)

    return df


def one_hot_encoder(df, input_cols, output_col=None, **kargs):
    """
    Maps a column of label indices to a column of binary vectors, with at most a single one-value.
    :param df: Dataframe to be transformed.
    :param input_cols: Columns to be encoded.
    :param output_col: Column where the output is going to be saved.
    :return: Dataframe with encoded columns.
    """

    input_cols = parse_columns(df, input_cols)

    if output_col is None:
        output_col = name_col(input_cols, "one_hot_encoder")

    encode = [OneHotEncoder(inputCol=column, outputCol=output_col, **kargs) for column in
              list(set(input_cols))]

    pipeline = Pipeline(stages=encode)
    df = pipeline.fit(df).transform(df)

    return df


# TODO: Must we use the pipeline version?
def vector_assembler(df, input_cols, output_col=None):
    """
    Combines a given list of columns into a single vector column.
    :param df: Dataframe to be transformed.
    :param input_cols: Columns to be assembled.
    :param output_col: Column where the output is going to be saved.
    :return: Dataframe with assembled column.
    """

    input_cols = parse_columns(df, input_cols)

    if output_col is None:
        output_col = name_col(input_cols, "vector_assembler")

    assembler = [VectorAssembler(inputCols=input_cols, outputCol=output_col)]

    pipeline = Pipeline(stages=assembler)
    df = pipeline.fit(df).transform(df)

    return df


def normalizer(df, input_cols, output_col=None, p=2.0):
    """
    Transforms a dataset of Vector rows, normalizing each Vector to have unit norm. It takes parameter p, which
    specifies the p-norm used for normalization. (p=2) by default.
    :param df: Dataframe to be transformed
    :param input_cols: Columns to be normalized.
    :param output_col: Column where the output is going to be saved.
    :param p:  p-norm used for normalization.
    :return: Dataframe with normalized columns.
    """

    # Check if columns argument must be a string or list datat ype:
    if not is_(input_cols, (str, list)):
        RaiseIt.type_error(input_cols, ["str", "list"])

    if is_str(input_cols):
        input_cols = [input_cols]

    if is_(input_cols, (float, int)):
        RaiseIt.type_error(input_cols, ["float", "int"])

    # Try to create a vector
    if len(input_cols) > 1:
        df = df.cols.cast(input_cols, "vector")

    if output_col is None:
        output_col = name_col(input_cols, "normalizer")

    # TODO https://developer.ibm.com/code/2018/04/10/improve-performance-ml-pipelines-wide-dataframes-apache-spark-2-3/
    normal = [Normalizer(inputCol=col_name, outputCol=output_col, p=p) for col_name in
              list(set(input_cols))]

    pipeline = Pipeline(stages=normal)

    df = pipeline.fit(df).transform(df)

    return df



from optimus.engines.base.commons.functions import is_string, is_integer, is_float, is_numeric
from optimus.engines.base.mask import Mask


class DaskMask(Mask):

    def str(self, cols="*"):
        return self.root[cols].cols.apply(cols, is_string, mode="partitioned")

    def int(self, cols="*"):
        return self.root[cols].cols.apply(cols, is_integer, mode="partitioned")

    def float(self, cols="*"):
        return self.root[cols].cols.apply(cols, is_float, mode="partitioned")

    def numeric(self, cols="*"):
        return self.root[cols].cols.apply(cols, is_numeric, mode="partitioned")



import dask
import dask.dataframe as dd
import pandas as pd
from dask import delayed
from sklearn.preprocessing import MinMaxScaler

from optimus.engines.base.meta import Meta
from optimus.helpers.core import val_to_list
from optimus.helpers.columns import parse_columns, get_output_cols, name_col
from optimus.helpers.constants import Actions
from optimus.infer import is_list_value
from optimus.profiler.functions import fill_missing_var_types


class DaskBaseColumns():

    @staticmethod
    def exec_agg(exprs, compute):
        """
        Execute and aggregation
        :param exprs:
        :return:
        """

        if compute:
            result = dask.compute(exprs)[0]
            if is_list_value(exprs):
                result = result[0].to_dict()
        else:
            result = delayed(exprs)

        return result

    def append(self, dfs):
        """

        :param dfs:
        :return:
        """

        dfs = val_to_list(dfs)

        df = self.root
        dfd = dd.concat([df.data.reset_index(drop=True), *[_df.data.reset_index(drop=True) for _df in dfs]], axis=1)
        meta = Meta.action(df.meta, Actions.APPEND.value, df.cols.names())
        return self.root.new(dfd, meta=meta)
        # return dfd

    def qcut(self, columns, quantiles, handle_invalid="skip"):

        df = self.root.data
        columns = parse_columns(df, columns)
        # s.fillna(np.nan)
        df[columns] = df[columns].map_partitions(pd.qcut, quantiles)
        return df

    @staticmethod
    def heatmap(columns, buckets=10):
        pass

    @staticmethod
    def standard_scaler():
        pass

    @staticmethod
    def max_abs_scaler(input_cols, output_cols=None):
        pass

    def min_max_scaler(self, input_cols, output_cols=None):
        # https://github.com/dask/dask/issues/2690

        df = self.df

        scaler = MinMaxScaler()

        input_cols = parse_columns(df, input_cols)
        output_cols = get_output_cols(input_cols, output_cols)

        # _df = df[input_cols]
        scaler.fit(df[input_cols])
        # print(type(scaler.transform(_df)))
        arr = scaler.transform(df[input_cols])
        darr = dd.from_array(arr)
        # print(type(darr))
        darr.name = 'z'
        df = df.merge(darr)

        return df

    # Date operations
    @staticmethod
    def to_timestamp(input_cols, date_format=None, output_cols=None):
        pass

    def replace_regex(self, input_cols, regex=None, value="", output_cols=None):
        """
        Use a Regex to replace values
        :param input_cols: '*', list of columns names or a single column name.
        :param output_cols:
        :param regex: values to look at to be replaced
        :param value: new value to replace the old one
        :return:
        """

        def _replace_regex(_value, _regex, _replace):
            return _value.replace(_regex, _replace, regex=True)

        return self.apply(input_cols, func=_replace_regex, args=(regex, value,), output_cols=output_cols,
                          filter_col_by_dtypes=self.root.constants.STRING_TYPES + self.root.constants.NUMERIC_TYPES)

    def reverse(self, input_cols, output_cols=None):
        def _reverse(value):
            return value.astype(str).str[::-1]

        return self.apply(input_cols, _reverse, output_cols=output_cols, mode="pandas", set_index=True)

    @staticmethod
    def astype(*args, **kwargs):
        pass

    # TODO: Check if we must use * to select all the columns

    def count_by_dtypes(self, columns, infer=False, str_funcs=None, int_funcs=None, mismatch=None):
        df = self.root.data
        columns = parse_columns(df, columns)
        columns_dtypes = df.cols.dtypes()

        def value_counts(series):
            return series.value_counts()

        delayed_results = []

        for col_name in columns:
            a = df.map_partitions(lambda df: df[col_name].apply(
                lambda row: Infer.parse((col_name, row), infer, columns_dtypes, str_funcs, int_funcs,
                                        full=False))).compute()

            f = df.functions.map_delayed(a, value_counts)
            delayed_results.append({col_name: f.to_dict()})

        results_compute = dask.compute(*delayed_results)
        result = {}

        # Convert list to dict
        for i in results_compute:
            result.update(i)

        if infer is True:
            result = fill_missing_var_types(result, columns_dtypes)
        else:
            result = self.parse_profiler_dtypes(result)

        return result

    def nest(self, input_cols, separator="", output_col=None, drop=False, shape="string"):
        """
        Merge multiple columns with the format specified
        :param input_cols: columns to be nested
        :param separator: char to be used as separator at the concat time
        :param shape: final data type, 'array', 'string' or 'vector'
        :param output_col:
        :return: Dask DataFrame
        """

        df = self.root
        input_cols = parse_columns(df, input_cols)
        # output_col = val_to_list(output_col)
        # check_column_numbers(input_cols, 2)
        if output_col is None:
            output_col = name_col(input_cols)
            # RaiseIt.type_error(output_col, ["str"])

        # output_col = parse_columns(df, output_col, accepts_missing_cols=True)

        output_ordered_columns = df.cols.names()

        def _nest_string(row):
            v = row[input_cols[0]].astype(str)
            for i in range(1, len(input_cols)):
                v = v + separator + row[input_cols[i]].astype(str)
            return v

        def _nest_array(row):
            # https://stackoverflow.com/questions/43898035/pandas-combine-column-values-into-a-list-in-a-new-column/43898233
            # t['combined'] = t.values.tolist()

            v = row[input_cols[0]].astype(str)
            for i in range(1, len(input_cols)):
                v += ", " + row[input_cols[i]].astype(str)
            return "[" + v + "]"

        if shape == "string":
            kw_columns = {output_col: _nest_string}
        else:
            kw_columns = {output_col: _nest_array}

        dfd = df.cols.assign(kw_columns).data

        if output_col not in output_ordered_columns:
            col_index = output_ordered_columns.index(input_cols[-1]) + 1
            output_ordered_columns[col_index:col_index] = [output_col]

        meta = Meta.action(df.meta, Actions.NEST.value, list(kw_columns.values()))

        if drop is True:
            for input_col in input_cols:
                if input_col in output_ordered_columns and input_col != output_col:
                    output_ordered_columns.remove(input_col)

        return self.root.new(dfd, meta).cols.select(output_ordered_columns)



# This functions must handle one or multiple columns
# Must return None if the data type can not be handle


from datetime import datetime, timedelta

import dask
import dask.array as da
import pandas as pd
from dask.array import stats

from optimus.engines.base.commons.functions import to_float, to_integer, to_boolean, to_datetime, word_tokenize
from optimus.engines.base.functions import Functions
from optimus.engines.base.dask.functions import DaskBaseFunctions
from optimus.helpers.core import val_to_list


class DaskFunctions(DaskBaseFunctions, Functions):

    @property
    def constants(self):
        from optimus.engines.base.dask.constants import Constants
        return Constants()

    def delayed(self, func):
        def wrapper(*args, **kwargs):
            return dask.delayed(func)(*args, **kwargs)

        return wrapper

    def to_float(self, series):
        return to_float(series)

    def to_integer(self, series):
        return to_integer(series)

    def to_boolean(self, series):
        return to_boolean(series)

    def to_string(self, series):
        if str(series.dtype) in self.constants.STRING_TYPES:
            return series
        return series.astype(str)

    def word_tokenize(self, value):
        return word_tokenize(value)

    def kurtosis(self, series):
        return stats.kurtosis(self._to_float(series))

    def skew(self, series):
        return stats.skew(self._to_float(series))

    def exp(self, series):
        return da.exp(self._to_float(series))

    def sqrt(self, series):
        return da.sqrt(self._to_float(series))

    def unique_values(self, series, *args):
        # print("args",args)
        # Cudf can not handle null so we fill it with non zero values.
        return self.to_string(series).unique()

    def radians(self, series):
        return da.radians(self._to_float(series))

    def degrees(self, series):
        return da.degrees(self._to_float(series))

    def ln(self, series):
        return da.log(self._to_float(series))

    def log(self, series, base=10):
        return da.log(self._to_float(series)) / da.log(base)

    def ceil(self, series):
        return da.ceil(self._to_float(series))

    def floor(self, series):
        return da.floor(self._to_float(series))

    def sin(self, series):
        return da.sin(self._to_float(series))

    def cos(self, series):
        return da.cos(self._to_float(series))

    def tan(self, series):
        return da.tan(self._to_float(series))

    def asin(self, series):
        return da.arcsin(self._to_float(series))

    def acos(self, series):
        return da.arccos(self._to_float(series))

    def atan(self, series):
        return da.arctan(self._to_float(series))

    def sinh(self, series):
        return da.arcsinh(self._to_float(series))

    def cosh(self, series):
        return da.cosh(self._to_float(series))

    def tanh(self, series):
        return da.tanh(self._to_float(series))

    def asinh(self, series):
        return da.arcsinh(self._to_float(series))

    def acosh(self, series):
        return da.arccosh(self._to_float(series))

    def atanh(self, series):
        return da.arctanh(self._to_float(series))

    def normalize_chars(self, series):
        # str.decode return a float column. We are forcing to return a string again
        return series.str.normalize("NFKD").str.encode('ascii', errors='ignore').str.decode('utf8').astype(str)

    def date_format(self, series, current_format=None, output_format=None):
        return pd.to_datetime(series, format=current_format, errors="coerce").dt.strftime(output_format)

    def days_between(self, series, date_format=None):
        return (pd.to_datetime(series, format=date_format,
                               errors="coerce").dt.date - datetime.now().date())

    def to_datetime(self, series, format):
        return to_datetime(series, format)



from optimus.helpers.types import DataFrameType


class BaseSet():
    """Base class for all set implementations"""

    def __init__(self, root):
        self.root = root

    def __call__(self, cols=None, value=None, default=None, args=[], where=None):
        return self.root.cols.set(cols=cols, value=value, args=args, where=where, default=default)


    def all(self, cols, value=None) -> DataFrameType:
        return self.__call__(cols, value)


    def _mask(self, cols, true_value=None, default_value=None, args=[], method: str=None, **kwargs) -> DataFrameType:
        
        df = self.root
        
        if cols in df.cols.names():
            input = df[cols]
            input_cols = cols
        elif isinstance(true_value, (self.root.__class__,)):
            input = true_value
            input_cols = input.cols.names()[0]
        else:
            df[cols] = default_value or true_value
            input = df[cols]
            input_cols = cols

        mask = getattr(input.mask, method)(input_cols, **kwargs)
        return self.__call__(cols, value=true_value, default=default_value, args=args, where=mask)

    # Types

    def str(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="str")

    def int(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="int")
    
    def float(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="float")
    
    def numeric(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="numeric")
    
    def email(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="email")
    
    def ip(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="ip")
    
    def url(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="url")
    
    def gender(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="gender")
    
    def boolean(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="boolean")
    
    def zip_code(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="zip_code")
    
    def credit_card_number(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="credit_card_number")
    
    def datetime(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="datetime")
    
    def object(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="object")
    
    def array(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="array")
    
    def phone_number(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="phone_number")
    
    def social_security_number(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="social_security_number")
    
    def http_code(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="http_code")

    # Other    

    def null(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="null")

    def none(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="none")
    
    def nan(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="nan")
    
    def empty(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="empty")

    def greater_than(self, cols, true_value=None, default_value=None, args=[], value=None) -> DataFrameType:
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, method="greater_than", value=value)

    def greater_than_equal(self, cols, true_value=None, default_value=None, args=[], value=None) -> DataFrameType:
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, method="greater_than_equal", value=value)

    def less_than(self, cols, true_value=None, default_value=None, args=[], value=None) -> DataFrameType:
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, method="less_than", value=value)

    def less_than_equal(self, cols, true_value=None, default_value=None, args=[], value=None) -> DataFrameType:
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, method="less_than_equal", value=value)

    def equal(self, cols, true_value=None, default_value=None, args=[], value=None) -> DataFrameType:
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, method="equal", value=value)

    def not_equal(self, cols, true_value=None, default_value=None, args=[], value=None) -> DataFrameType:
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, method="not_equal", value=value)

    def missing(self, cols, value=None, default_value=None, args=[]) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="missing")

    def mismatch(self, cols, value=None, default_value=None, args=[], dtype=None) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="mismatch", dtype=dtype)

    def duplicated(self, cols, value=None, default_value=None, args=[], keep="first") -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="duplicated", keep=keep)

    def unique(self, cols, value=None, default_value=None, args=[], keep="first") -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="unique", keep=keep)

    def match(self, cols, value=None, default_value=None, args=[], regex=None, dtype=None) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="match_dtype", regex=regex, dtype=dtype)

    def match_regex(self, cols, value=None, default_value=None, args=[], regex=None) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="match_regex", regex=regex)

    def match_dtype(self, cols, value=None, default_value=None, args=[], dtype=None) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="match_dtype", dtype=dtype)

    def match_pattern(self, cols, value=None, default_value=None, args=[], pattern=None) -> DataFrameType:
        return self._mask(cols, true_value=value, default_value=default_value, args=args, method="match_pattern", pattern=pattern)

    def value_in(self, cols, true_value=None, default_value=None, args=[], values=None) -> DataFrameType:
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, method="value_in", values=values)
        
    def starts_with(self, cols, true_value=None, default_value=None, args=[], value=None) -> DataFrameType:
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, method="starts_with", value=value)

    def ends_with(self, cols, true_value=None, default_value=None, args=[], value=None) -> DataFrameType:
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, method="ends_with", value=value)

    def contains(self, cols, true_value=None, default_value=None, args=[], value=None) -> DataFrameType:
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, method="contains", value=value)
    
    def expression(self, cols, true_value=None, default_value=None, args=[], where=None) -> DataFrameType:
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, method="expression", where=where)




import base64
import simplejson as json
import os
import pprint
from io import BytesIO

from IPython.core.display import display, HTML
from matplotlib import pyplot as plt

from optimus.infer import is_str


def output_image(fig, path):
    """
    Output a png file
    :param fig:
    :param path: Matplotlib figure
    :return: Base64 encode image
    """

    fig.savefig(path, format='png')
    plt.close()


def output_base64(fig):
    """
    Output a matplotlib as base64 encode
    :param fig: Matplotlib figure
    :return: Base64 encode image
    """
    fig_file = BytesIO()
    plt.savefig(fig_file, format='png')
    # rewind to beginning of file
    fig_file.seek(0)

    fig_png = base64.b64encode(fig_file.getvalue())
    plt.close(fig)

    return fig_png.decode('utf8')


def print_html(html):
    """
    Display() helper to print html code
    :param html: html code to be printed
    :return:
    """
    try:
        if "DATABRICKS_RUNTIME_VERSION" in os.environ:
            displayHTML(result)
        else:
            display(HTML(html))
        return True
    except NameError:
        return False


def print_json(value):
    """
    Print a human readable json
    :param value: json to be printed
    :return: json
    """
    pp = pprint.PrettyPrinter(indent=2)
    if is_str(value):
        value = value.replace("'", "\"")
        value = json.loads(value)

    pp.pprint(value)



import cudf
from dask import dataframe as dd

from optimus.engines.base.dask.rows import DaskBaseRows
from optimus.infer import is_list_value


class Rows(DaskBaseRows):

    def __init__(self, df):
        super(DaskBaseRows, self).__init__(df)

    def append(self, rows):
        """

        :param rows:
        :return:
        """
        df = self.df

        if is_list_value(rows):
            rows = dd.from_pandas(cudf.DataFrame(rows), npartitions=1)

        # Can not concatenate dataframe with not string columns names
        rows.columns = df.columns

        df = dd.concat([df, rows], axis=0, interleave_partitions=True)

        return df



import functools
import operator

import cudf
import pandas as pd
from cudf.core import DataFrame
from multipledispatch import dispatch

from optimus.engines.base.rows import BaseRows
from optimus.helpers.columns import parse_columns
from optimus.helpers.constants import Actions
from optimus.helpers.core import val_to_list, one_list_to_val
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_list_of_str_or_int, is_list_value
from optimus.engines.base.meta import Meta


class Rows(BaseRows):
    def __init__(self, df):
        super(Rows, self).__init__(df)

    @staticmethod
    def create_id(column="id") -> DataFrame:
        pass

    def append(self, rows):
        """

        :param rows:
        :return:
        """
        df = self.root

        if is_list_value(rows):
            rows = cudf.DataFrame(rows)
        # Can not concatenate dataframe with not string columns names

        rows.columns = df.cols.names()
        # df = cudf.concat([df.reset_index(drop=True), rows.reset_index(drop=True)], axis=0)
        df = cudf.concat([df, rows], axis=0, ignore_index=True)
        return df

    def _sort(self, dfd, col_name, ascending):
        return dfd.sort_values(col_name, ascending=ascending)

    @staticmethod
    def unnest(input_cols) -> DataFrame:
        df = self.root
        return df



def val_to_list(val):
    """
    Convert a single value string or number to a list
    :param val:
    :return:
    """
    if val is not None:
        if not isinstance(val, (list, tuple)):
            val = [val]
        if isinstance(val, tuple):
            val = list(val)
    return val


def one_list_to_val(val):
    """
    Convert a single list element to val
    :param val:
    :return:
    """
    if isinstance(val, list) and len(val) == 1:
        result = val[0]
    else:
        result = val

    return result



"""
Code based on pyspark_pipes by Daniel Acuña
"""
from pyspark.ml import Pipeline
from pyspark.ml.param.shared import HasFeaturesCol, HasInputCol, \
    HasInputCols, HasPredictionCol, HasOutputCol, HasLabelCol, Params, \
    HasRawPredictionCol, HasProbabilityCol

ALLOWED_TYPES = (HasFeaturesCol, HasInputCol, HasInputCols, HasLabelCol,
                 HasPredictionCol, HasOutputCol)


def set_default_colnames(pipe_stage):
    """
    Add object ids to default columns
    :param pipe_stage: Params object
    :return: modified stage
    """
    # pylint: disable=protected-access
    if isinstance(pipe_stage, HasFeaturesCol) and not pipe_stage.isSet('featuresCol'):
        pipe_stage._setDefault(featuresCol=pipe_stage.uid + '__features')
    if isinstance(pipe_stage, HasRawPredictionCol) and not pipe_stage.isSet('rawPredictionCol'):
        pipe_stage._setDefault(rawPredictionCol=pipe_stage.uid + '__rawPrediction')
    if isinstance(pipe_stage, HasProbabilityCol) and not pipe_stage.isSet('probabilityCol'):
        pipe_stage._setDefault(probabilityCol=pipe_stage.uid + '__probability')
    if isinstance(pipe_stage, HasPredictionCol) and not pipe_stage.isSet('predictionCol'):
        pipe_stage._setDefault(predictionCol=pipe_stage.uid + '__prediction')
    return pipe_stage


def is_instance(obj, typelist):
    """
    Returns true if object `obj` is of any of the types list in `typelist`
    :param obj: object
    :param typelist: list of types
    :return: bool
    """
    return any([isinstance(obj, t) for t in typelist])


def get_pipeline_laststep(pipe):
    """
    Get last stage of pipe that is not a Pipeline
    :param pipe: Params
    :return: Params
    """
    if isinstance(pipe, Pipeline):
        return get_pipeline_laststep(pipe.getStages()[-1])
    return pipe


def get_pipeline_firststep(pipe):
    """
    Get last stage of a pipe that is not a Pipeline
    :param pipe: Params
    :return: Estimator or Transformer
    """
    if isinstance(pipe, Pipeline):
        return get_pipeline_firststep(pipe.getStages()[0])

    return pipe


def left_pipe_function(self, other):
    """
    Infix pipe operator
    :param self: right-hand side object
    :param other: left-hand side object
    :return: Pipeline
    """
    # pylint: disable=too-many-branches

    self = set_default_colnames(self)
    # At the start of the Pipeline we need to put together two things
    if is_instance(other, ALLOWED_TYPES):
        result = Pipeline().setStages([other]) | self
    elif isinstance(other, Pipeline):
        last_step = set_default_colnames(get_pipeline_laststep(other))
        if isinstance(last_step, HasOutputCol):
            # let's generate some random string to represent the column's name
            last_step_output = last_step.getOutputCol()
        elif isinstance(last_step, HasPredictionCol):
            last_step_output = last_step.getPredictionCol()
        else:
            raise Exception("Type of step not supported")

        # should we connect input with output?
        first_step = set_default_colnames(get_pipeline_firststep(self))
        if isinstance(first_step, HasInputCol):
            if not first_step.isSet('inputCol'):
                first_step.setInputCol(last_step_output)
        if isinstance(first_step, HasFeaturesCol):
            if not first_step.isSet('featuresCol'):
                first_step.setFeaturesCol(last_step_output)

        result = Pipeline().setStages(other.getStages() + [self])
    elif isinstance(other, (tuple, list)):
        # check that connecting to one estimator or transformer
        if not isinstance(self, HasInputCols):
            raise Exception("When many to one connection, then receiver "
                            "must accept multiple inputs")

        all_outputs = []
        all_objects = []
        for stage in other:
            if not isinstance(stage, NotBroadcasted):
                last_step = set_default_colnames(get_pipeline_laststep(stage))

                if isinstance(last_step, HasOutputCol):
                    # let's generate some random string to represent the column's name
                    last_step_output = last_step.getOutputCol()
                elif isinstance(last_step, HasPredictionCol):
                    last_step_output = last_step.getPredictionCol()
                else:
                    raise Exception("It must contain output or prediction")

                all_outputs.append(last_step_output)
                all_objects.append(stage)
            else:
                all_objects.append(stage.object)

        # should we connect input with output?
        first_step = set_default_colnames(get_pipeline_firststep(self))
        if not first_step.isSet('inputCols'):
            first_step.setInputCols(all_outputs)

        result = Pipeline().setStages(all_objects + [self])
    else:
        raise Exception("Type of pipeline not supported")

    return result


def right_pipe_function(self, other):
    """
    Prefix pipe operator
    :param self: left-hand side operator
    :param other: right-hand side operator
    :return: Pipeline
    """
    self = set_default_colnames(self)
    if isinstance(other, (list, tuple)) and \
            isinstance(self, tuple(list(ALLOWED_TYPES) + [Pipeline])):
        last_step = set_default_colnames(get_pipeline_laststep(self))

        if isinstance(last_step, HasOutputCol):
            # let's generate some random string to represent the column's name
            last_step_output = last_step.getOutputCol()
        elif isinstance(last_step, HasPredictionCol):
            last_step_output = last_step.getPredictionCol()
        else:
            raise Exception("It must contain output or prediction")

        for stage in other:
            first_step = set_default_colnames(get_pipeline_firststep(stage))

            if isinstance(first_step, HasInputCol):
                # let's generate some random string to represent the column's name
                if not first_step.isSet('inputCol'):
                    first_step.setInputCol(last_step_output)
            elif isinstance(first_step, HasFeaturesCol):
                if not first_step.isSet('featuresCol'):
                    first_step.setFeaturesCol(last_step_output)
            else:
                raise Exception("A step didn't allow inputs")

        result = [NotBroadcasted(self)] + list(other)
    else:
        result = left_pipe_function(other, self)
    return result


class NotBroadcasted:
    """
    Wraps a Params object so that it doesn't broadcast it as part of a many to one
    pipe transformation
    """

    # pylint: disable=too-few-public-methods
    def __init__(self, params_object):
        self.object = params_object


def patch():
    """
    Monkey patches the Params class
    :return: None
    """
    Params.__or__ = right_pipe_function
    Params.__ror__ = left_pipe_function


def print_stage(pipe):
    """
    Print stage of a pipeline even recursively
    :param pipe: Pipeline
    :return: str
    """
    if isinstance(pipe, Pipeline):
        return "[\n" + ','.join([print_stage(s) for s in pipe.getStages()]) + "\n]"
    else:
        result = ""
        if isinstance(pipe, HasInputCol):
            result += pipe.getInputCol()
        elif isinstance(pipe, HasInputCols):
            result += str(pipe.getInputCols())
        elif isinstance(pipe, HasFeaturesCol):
            result += pipe.getFeaturesCol()

        result += " - "
        if isinstance(pipe, HasOutputCol):
            result += pipe.getOutputCol()
        elif isinstance(pipe, HasPredictionCol):
            result += pipe.getPredictionCol()
        return result



from optimus.helpers.types import ModelType

class ML:
    def __init__(self, root):
        self.root = root

    def linear_regression(self, features, target, *args, **kwargs) -> ModelType:
        pass

    def logistic_regression(self, features, target, *args, **kwargs) -> ModelType:
        pass

    def KMeans(self, features, n_centers, *args, **kwargs) -> ModelType:
        pass

    def logistic_regression_text(self, input_col) -> ModelType:
        pass

    def decision_tree(self, columns, input_col, **kwargs) -> ModelType:
        pass

    def random_forest(self, features, target, *args, **kwargs) -> ModelType:
        pass

    def gbt(self, columns, input_col, **kwargs) -> ModelType:
        pass



import logging
import sys
from datetime import date, datetime

from pyspark.sql.types import *

from optimus import Optimus

op = Optimus()

s_logger = logging.getLogger('py4j.java_gateway')
s_logger.setLevel(logging.INFO)

source_df = op.create.df(
    [
        ("names", "str", True),
        ("height(ft)", "int", True),
        ("function", "str", True),
        ("rank", "int", True),
        ("age", "int", True),
        ("weight(t)", "float", True),
        ("japanese name", ArrayType(StringType()), True),
        ("last position seen", "str", True),
        ("date arrival", "str", True),
        ("last date seen", "str", True),
        ("attributes", ArrayType(FloatType()), True),
        ("DateType", DateType()),
        ("Timestamp", TimestampType()),
        ("Cybertronian", "bool", True),
        ("function(binary)", "binary", False),
        ("NullType", "null", True),

    ],
    [
        ("Optim'us", 28, "Leader", 10, 5000000, 4.30, ["Inochi", "Convoy"], "19.442735,-99.201111", "1980/04/10",
         "2016/09/10", [8.5344, 4300.0], date(2016, 9, 10), datetime(2014, 6, 24), True, bytearray("Leader", "utf-8"),
         None),
        ("bumbl#ebéé  ", 17, "Espionage", 7, 5000000, 2.0, ["Bumble", "Goldback"], "10.642707,-71.612534", "1980/04/10",
         "2015/08/10", [5.334, 2000.0], date(2015, 8, 10), datetime(2014, 6, 24), True, bytearray("Espionage", "utf-8"),
         None),
        ("ironhide&", 26, "Security", 7, 5000000, 4.0, ["Roadbuster"], "37.789563,-122.400356", "1980/04/10",
         "2014/07/10", [7.9248, 4000.0], date(2014, 6, 24), datetime(2014, 6, 24), True, bytearray("Security", "utf-8"),
         None),
        ("Jazz", 13, "First Lieutenant", 8, 5000000, 1.80, ["Meister"], "33.670666,-117.841553", "1980/04/10",
         "2013/06/10", [3.9624, 1800.0], date(2013, 6, 24), datetime(2014, 6, 24), True,
         bytearray("First Lieutenant", "utf-8"), None),
        ("Megatron", None, "None", 10, 5000000, 5.70, ["Megatron"], None, "1980/04/10", "2012/05/10", [None, 5700.0],
         date(2012, 5, 10), datetime(2014, 6, 24), True, bytearray("None", "utf-8"), None),
        ("Metroplex_)^$", 300, "Battle Station", 8, 5000000, None, ["Metroflex"], None, "1980/04/10", "2011/04/10",
         [91.44, None], date(2011, 4, 10), datetime(2014, 6, 24), True, bytearray("Battle Station", "utf-8"), None),
        ("1", 2, "3", 4, 5, 6.0, ["7"], 8, "1980/04/10", "2011/04/10",
         [11.0], date(2011, 4, 10), datetime(2014, 6, 24), True, bytearray("15", "utf-8"), None)
    ], infer_schema=True)


class TestProfiler(object):
    @staticmethod
    def test_run():

        try:
            op.profiler.run(source_df, "*")

        except RuntimeError:
            logging.exception('Error executing the profiler.')
            sys.exit(1)


def test_save_to_html():
    try:
        op.profiler.run(source_df, "*").to_file("output.html", "html")

    except RuntimeError:
        logging.exception('Error creating the html output.')
        sys.exit(1)


def test_save_to_json():
    try:
        op.profiler.run(source_df, "*").to_file("output.json", "json")

    except RuntimeError:
        logging.exception('Error creating the json output.')
        sys.exit(1)



from pyspark.sql import Row, types
from pyspark.ml import feature, classification
from nose.tools import assert_equal
import pyspark
import py_sparkling

from optimus import Optimus

import optimus.engines.spark.ml.encoding as fe

op = Optimus(master='local')


df_cancer = op.spark.read.csv('tests/data_cancer.csv', sep=',', header=True, inferSchema=True)
columns = ['diagnosis', 'radius_mean', 'texture_mean', 'perimeter_mean', 'area_mean', 'smoothness_mean',
           'compactness_mean', 'concavity_mean', 'concave points_mean', 'symmetry_mean',
           'fractal_dimension_mean']

columns_h2o = ['radius_mean', 'texture_mean', 'perimeter_mean', 'area_mean', 'smoothness_mean',
           'compactness_mean', 'concavity_mean', 'concave points_mean', 'symmetry_mean',
           'fractal_dimension_mean']


def assert_spark_df(df):
    assert isinstance(df, pyspark.sql.dataframe.DataFrame), "Not a Spark DF"


def assert_spark_model(model):
    assert isinstance(model, pyspark.ml.PipelineModel), "Not a model"


def test_ml_pipe():
    df = op.sc. \
         parallelize([Row(sentence='this is a test', label=0.),
                     Row(sentence='this is another test', label=1.)]). \
         toDF()

    pl = feature.Tokenizer().setInputCol('sentence') | feature.CountVectorizer()
    ml = pl | classification.LogisticRegression()

    ml_model = ml.fit(df)
    assert_equal(ml_model.transform(df).count(), 2)


def test_logistic_regression_text():
    df = op.sc. \
        parallelize([Row(sentence='this is a test', label=0.),
                     Row(sentence='this is another test', label=1.)]). \
        toDF()

    df_predict, ml_model = op.ml.logistic_regression_text(df, "sentence")

    assert_spark_df(df_predict)

    assert_spark_model(ml_model)


def test_n_gram():
    df = op.sc. \
        parallelize([['this is the best sentence ever'],
                     ['this is however the worst sentence available']]). \
        toDF(schema=types.StructType().add('sentence', types.StringType()))

    df_model, tfidf_model = fe.n_gram(df, input_col="sentence", n=2)

    assert_spark_df(df_model)

    assert_spark_model(tfidf_model)

    assert_equal(df_model.select('sentence', 'features').count(), 2)


def test_string_to_index_kargs():
    df = op.spark.createDataFrame([(0, "a"), (1, "b"), (2, "c"), (3, "a"), (4, "a"), (5, "c")],
                                 ["id", "category"])

    df_indexed = fe.string_to_index(df, "category", stringOrderType="frequencyAsc")

    assert_spark_df(df_indexed)

    expected_collect = op.sc.parallelize([Row(id=0, category='a', category_index=2.0),
                                          Row(id=1, category='b', category_index=0.0),
                                          Row(id=2, category='c', category_index=1.0),
                                          Row(id=3, category='a', category_index=2.0),
                                          Row(id=4, category='a', category_index=2.0),
                                          Row(id=5, category='c', category_index=1.0)]).toDF()

    assert_equal(df_indexed.select("category", "category***STRING_TO_INDEX", "id").collect(), expected_collect.collect())


def test_random_forest():
    df_model, rf_model = op.ml.random_forest(df_cancer, columns, "diagnosis")

    assert_spark_df(df_model)

    assert isinstance(rf_model, pyspark.ml.classification.RandomForestClassificationModel), "Not a RF model"


def test_decision_tree():
    df_model, rf_model = op.ml.decision_tree(df_cancer, columns, "diagnosis")

    assert_spark_df(df_model)

    assert isinstance(rf_model, pyspark.ml.classification.DecisionTreeClassificationModel), "Not a DT model"


def test_gbt():
    df_model, rf_model = op.ml.gbt(df_cancer, columns, "diagnosis")

    assert_spark_df(df_model)

    assert isinstance(rf_model, pyspark.ml.classification.GBTClassificationModel), "Not a GBT model"


def test_h2o_automl():
    df_model, automl_model = op.ml.h2o_automl(df_cancer, "diagnosis", columns_h2o)

    assert_spark_df(df_model)

    assert isinstance(automl_model, py_sparkling.ml.models.H2OMOJOModel), "Not a H2OMOJOModel"


def test_h2o_deeplearning():
    df_model, dl_model = op.ml.h2o_deeplearning(df_cancer, "diagnosis", columns_h2o)

    assert_spark_df(df_model)

    assert isinstance(dl_model, py_sparkling.ml.models.H2OMOJOModel), "Not a H2OMOJOModel"


def test_h2o_xgboost():
    df_model, xgboost_model = op.ml.h2o_xgboost(df_cancer, "diagnosis", columns_h2o)

    assert_spark_df(df_model)

    assert isinstance(xgboost_model, py_sparkling.ml.models.H2OMOJOModel), "Not a H2OMOJOModel"


def test_h2o_gbm():
    df_model, gbm_model = op.ml.h2o_gbm(df_cancer, "diagnosis", columns_h2o)

    assert_spark_df(df_model)

    assert isinstance(gbm_model, py_sparkling.ml.models.H2OMOJOModel), "Not a H2OMOJOModel"


from optimus.engines.base.dataframe.dataframe import DataFrameBaseDataFrame
from optimus.engines.base.basedataframe import BaseDataFrame
from optimus.engines.cudf.io.save import Save
from optimus.engines.pandas.dataframe import PandasDataFrame


class CUDFDataFrame(DataFrameBaseDataFrame, BaseDataFrame):

    def __init__(self, data):
        super().__init__(data)

    def _base_to_dfd(self, pdf, n_partitions):
        pass

    def to_optimus_pandas(self):
        return PandasDataFrame(self.root.data.to_pandas())

    @property
    def rows(self):
        from optimus.engines.cudf.rows import Rows
        return Rows(self)

    @property
    def cols(self):
        from optimus.engines.cudf.columns import Cols
        return Cols(self)

    @property
    def save(self):
        return Save(self)

    @property
    def mask(self):
        from optimus.engines.cudf.mask import CUDFMask
        return CUDFMask(self)

    @property
    def functions(self):
        from optimus.engines.cudf.functions import CUDFFunctions
        return CUDFFunctions()

    @property
    def constants(self):
        from optimus.engines.cudf.constants import Constants
        return Constants()

    def _buffer_window(self, input_cols, lower_bound, upper_bound):
        return PandasDataFrame(self.data[input_cols][lower_bound: upper_bound].to_pandas())

    def encoding(self):
        pass

    def get_buffer(self):
        return self

    def to_pandas(self):
        return self.data.to_pandas()

    def to_dict(self, cols="*", n=10, orient="list"):
        """
        Create a dict
        :param n:
        :param orient:
        :return:
        """

        if n=="all":
            series = self.cols.select(cols).to_pandas()
        else:
            series = self.buffer_window(cols, 0, n).data
            
        return series.to_pandas().to_dict(orient)

    def repartition(self, n=None, *args, **kwargs):
        return self.root



from pyspark.sql.types import *

import optimus.helpers.functions_spark
from optimus import Optimus
import unittest
import numpy as np
nan = np.nan
op = Optimus(master='local')
source_df=op.create.df([('words', StringType(), True),('num', IntegerType(), True),('animals', StringType(), True),('thing', StringType(), True),('second', IntegerType(), True),('filter', StringType(), True)], [('  I like     fish  ', 1, 'dog dog', 'housé', 5, 'a'), ('    zombies', 2, 'cat', 'tv', 6, 'b'), ('simpsons   cat lady', 2, 'frog', 'table', 7, '1'), (None, 3, 'eagle', 'glass', 8, 'c')])
class Test_df_rows(unittest.TestCase):
	maxDiff = None
	@staticmethod
	def test_rows_append():
		actual_df = optimus.helpers.functions_spark.append([('this is a word', 2, 'this is an animal', 'this is a thing', 64, 'this is a filter')])
		expected_df = op.create.df([('words', StringType(), True),('num', IntegerType(), True),('animals', StringType(), True),('thing', StringType(), True),('second', IntegerType(), True),('filter', StringType(), True)], [('  I like     fish  ', 1, 'dog dog', 'housé', 5, 'a'), ('    zombies', 2, 'cat', 'tv', 6, 'b'), ('simpsons   cat lady', 2, 'frog', 'table', 7, '1'), (None, 3, 'eagle', 'glass', 8, 'c'), ('this is a word', 2, 'this is an animal', 'this is a thing', 64, 'this is a filter')])
		assert (expected_df.collect() == actual_df.collect())
	@staticmethod
	def test_rows_between():
		actual_df =source_df.rows.between('second',6,8)
		expected_df = op.create.df([('words', StringType(), True),('num', IntegerType(), True),('animals', StringType(), True),('thing', StringType(), True),('second', IntegerType(), True),('filter', StringType(), True)], [('simpsons   cat lady', 2, 'frog', 'table', 7, '1')])
		assert (expected_df.collect() == actual_df.collect())
	@staticmethod
	def test_rows_between_equal():
		actual_df =source_df.rows.between('second',6,8,equal=True)
		expected_df = op.create.df([('words', StringType(), True),('num', IntegerType(), True),('animals', StringType(), True),('thing', StringType(), True),('second', IntegerType(), True),('filter', StringType(), True)], [('    zombies', 2, 'cat', 'tv', 6, 'b'), ('simpsons   cat lady', 2, 'frog', 'table', 7, '1'), (None, 3, 'eagle', 'glass', 8, 'c')])
		assert (expected_df.collect() == actual_df.collect())
	@staticmethod
	def test_rows_between_invert_equal():
		actual_df =source_df.rows.between('second',6,8,invert=True,equal=True)
		expected_df = op.create.df([('words', StringType(), True),('num', IntegerType(), True),('animals', StringType(), True),('thing', StringType(), True),('second', IntegerType(), True),('filter', StringType(), True)], [('  I like     fish  ', 1, 'dog dog', 'housé', 5, 'a'), ('    zombies', 2, 'cat', 'tv', 6, 'b'), (None, 3, 'eagle', 'glass', 8, 'c')])
		assert (expected_df.collect() == actual_df.collect())
	@staticmethod
	def test_rows_drop_by_dtypes():
		actual_df =source_df.rows.drop_by_dtypes('filter','integer')
		expected_df = op.create.df([('words', StringType(), True),('num', IntegerType(), True),('animals', StringType(), True),('thing', StringType(), True),('second', IntegerType(), True),('filter', StringType(), True)], [('  I like     fish  ', 1, 'dog dog', 'housé', 5, 'a'), ('    zombies', 2, 'cat', 'tv', 6, 'b'), (None, 3, 'eagle', 'glass', 8, 'c')])
		assert (expected_df.collect() == actual_df.collect())
	@staticmethod
	def test_rows_is_in():
		actual_df =source_df.rows.is_in('num',2)
		expected_df = op.create.df([('words', StringType(), True),('num', IntegerType(), True),('animals', StringType(), True),('thing', StringType(), True),('second', IntegerType(), True),('filter', StringType(), True)], [('    zombies', 2, 'cat', 'tv', 6, 'b'), ('simpsons   cat lady', 2, 'frog', 'table', 7, '1')])
		assert (expected_df.collect() == actual_df.collect())
	@staticmethod
	def test_rows_select_by_dtypes():
		actual_df =source_df.rows.select_by_dtypes('filter','integer')
		expected_df = op.create.df([('words', StringType(), True),('num', IntegerType(), True),('animals', StringType(), True),('thing', StringType(), True),('second', IntegerType(), True),('filter', StringType(), True)], [('simpsons   cat lady', 2, 'frog', 'table', 7, '1')])
		assert (expected_df.collect() == actual_df.collect())
	@staticmethod
	def test_rows_sort():
		actual_df =source_df.rows.sort('num','desc')
		expected_df = op.create.df([('words', StringType(), True),('num', IntegerType(), True),('animals', StringType(), True),('thing', StringType(), True),('second', IntegerType(), True),('filter', StringType(), True)], [(None, 3, 'eagle', 'glass', 8, 'c'), ('    zombies', 2, 'cat', 'tv', 6, 'b'), ('simpsons   cat lady', 2, 'frog', 'table', 7, '1'), ('  I like     fish  ', 1, 'dog dog', 'housé', 5, 'a')])
		assert (expected_df.collect() == actual_df.collect())



# -*- coding: utf-8 -*-
# + {}
from pyspark.sql.types import *
import datetime
import sys

from optimus import Optimus
# -

op = Optimus()

source_df = op.create.df(
    [('names', StringType(), True), ('height(ft)', ShortType(), True), ('function', StringType(), True),
     ('rank', ByteType(), True), ('age', IntegerType(), True), ('weight(t)', FloatType(), True),
     ('japanese name', ArrayType(StringType(), True), True), ('last position seen', StringType(), True),
     ('date arrival', StringType(), True), ('last date seen', StringType(), True),
     ('attributes', ArrayType(FloatType(), True), True), ('DateType', DateType(), True),
     ('timestamp', TimestampType(), True), ('Cybertronian', BooleanType(), True),
     ('function(binary)', BinaryType(), True), ('NullType', NullType(), True)], [("Optim'us", 28, 'Leader', 10, 5000000,
                                                                                  4.300000190734863,
                                                                                  ['Inochi', 'Convoy'],
                                                                                  '19.442735,-99.201111', '1980/04/10',
                                                                                  '2016/09/10',
                                                                                  [8.53439998626709, 4300.0],
                                                                                  datetime.date(2016, 9, 10),
                                                                                  datetime.datetime(2014, 6, 24, 0, 0),
                                                                                  True, bytearray(b'Leader'), None), (
                                                                                     'bumbl#ebéé  ', 17, 'Espionage', 7,
                                                                                     5000000, 2.0,
                                                                                     ['Bumble', 'Goldback'],
                                                                                     '10.642707,-71.612534',
                                                                                     '1980/04/10',
                                                                                     '2015/08/10',
                                                                                     [5.334000110626221, 2000.0],
                                                                                     datetime.date(2015, 8, 10),
                                                                                     datetime.datetime(2014, 6, 24, 0,
                                                                                                       0),
                                                                                     True, bytearray(b'Espionage'),
                                                                                     None), (
                                                                                     'ironhide&', 26, 'Security', 7,
                                                                                     5000000, 4.0, ['Roadbuster'],
                                                                                     '37.789563,-122.400356',
                                                                                     '1980/04/10',
                                                                                     '2014/07/10',
                                                                                     [7.924799919128418, 4000.0],
                                                                                     datetime.date(2014, 6, 24),
                                                                                     datetime.datetime(2014, 6, 24, 0,
                                                                                                       0),
                                                                                     True, bytearray(b'Security'),
                                                                                     None), (
                                                                                     'Jazz', 13, 'First Lieutenant', 8,
                                                                                     5000000, 1.7999999523162842,
                                                                                     ['Meister'],
                                                                                     '33.670666,-117.841553',
                                                                                     '1980/04/10', '2013/06/10',
                                                                                     [3.962399959564209, 1800.0],
                                                                                     datetime.date(2013, 6, 24),
                                                                                     datetime.datetime(2014, 6, 24, 0,
                                                                                                       0),
                                                                                     True,
                                                                                     bytearray(b'First Lieutenant'),
                                                                                     None), (
                                                                                     'Megatron', None, 'None', 10,
                                                                                     5000000,
                                                                                     5.699999809265137, ['Megatron'],
                                                                                     None,
                                                                                     '1980/04/10', '2012/05/10',
                                                                                     [None, 5700.0],
                                                                                     datetime.date(2012, 5, 10),
                                                                                     datetime.datetime(2014, 6, 24, 0,
                                                                                                       0),
                                                                                     True, bytearray(b'None'), None), (
                                                                                     'Metroplex_)^$', 300,
                                                                                     'Battle Station',
                                                                                     8, 5000000, None, ['Metroflex'],
                                                                                     None,
                                                                                     '1980/04/10', '2011/04/10',
                                                                                     [91.44000244140625, None],
                                                                                     datetime.date(2011, 4, 10),
                                                                                     datetime.datetime(2014, 6, 24, 0,
                                                                                                       0),
                                                                                     True, bytearray(b'Battle Station'),
                                                                                     None)])

source_df.display()


class TestDataFrameCols(object):
    @staticmethod
    def test_correlation():

        try:
            source_df.plot.correlation(["age", "rank"])

        except RuntimeError:
            logging.exception('Error creating the json output.')
            sys.exit(1)

    @staticmethod
    def test_boxplot():

        try:
            source_df.plot.box(["age", "weight(t)"])

        except RuntimeError:
            logging.exception('Error creating the json output.')
            sys.exit(1)

    @staticmethod
    def test_scatterplot():

        try:
            source_df.plot.scatter(["age", "weight(t)"])

        except RuntimeError:
            logging.exception('Error creating the json output.')
            sys.exit(1)

    @staticmethod
    def test_hist():

        try:
            source_df.plot.hist("age")

        except RuntimeError:
            logging.exception('Error creating the json output.')
            sys.exit(1)

    @staticmethod
    def test_frequency():

        try:
            source_df.plot.frequency("age")

        except RuntimeError:
            logging.exception('Error creating the json output.')
            sys.exit(1)

    @staticmethod
    def test_qqplot():

        try:
            source_df.plot.qqplot("age")

        except RuntimeError:
            logging.exception('Error creating the json output.')
            sys.exit(1)


from optimus.engines.base.commons.functions import is_integer, is_float, is_numeric, is_string
from optimus.engines.base.mask import Mask


class PandasMask(Mask):

    def str(self, cols="*"):
        return self.root[cols].cols.apply(cols, is_string) 

    def int(self, cols="*"):
        return self.root[cols].cols.apply(cols, is_integer)

    def float(self, cols="*"):
        return self.root[cols].cols.apply(cols, is_float)

    def numeric(self, cols="*"):
        return self.root[cols].cols.apply(cols, is_numeric)